"""
This module handles multiple API v2 Client sessions.
"""
from __future__ import annotations

import functools
from typing import Any
from typing import Callable
from typing import cast
from typing import TYPE_CHECKING
from typing import TypeVar

from . import Client
from ..events import ClientAddEvent
from ..events import ClientUpdateEvent
from ..events import Eventable
from ..exceptions import InvalidClientRequestedError
from ..models import OAuthToken
from .repository import BaseTokenRepository
from .repository import SimpleTokenRepository

if TYPE_CHECKING:
    from types import TracebackType
    from typing import Optional
    from typing import Type
    from typing import Union

__all__ = ("ClientStorage",)

F = TypeVar("F", bound=Callable[..., Any])


class ClientStorage(Eventable):
    r"""OAuth sessions manager.

    :param \**kwargs:
        See below

    :Keyword Arguments:
        * *token_repository* (``BaseTokenRepository``) --
            Optional, defaults to ``aiosu.v2.repository.SimpleTokenRepository()``
        * *client_secret* (``str``)
        * *client_id* (``int``)
        * *base_url* (``str``) --
            Optional, base API URL, defaults to "https://osu.ppy.sh"
        * *create_app_client* (``bool``) --
            Optional, whether to automatically create guest clients, defaults to True
    """

    __slots__ = (
        "_token_repository",
        "client_secret",
        "client_id",
        "base_url",
        "__create_app_client",
        "clients",
    )

    def __init__(self, **kwargs: Any) -> None:
        super().__init__()
        self._register_event(ClientAddEvent)
        self._register_event(ClientUpdateEvent)
        self._token_repository: BaseTokenRepository = kwargs.pop(
            "token_repository",
            SimpleTokenRepository(),
        )
        self.client_secret: str = kwargs.pop("client_secret", None)
        self.client_id: int = kwargs.pop("client_id", None)
        self.base_url: str = kwargs.pop("base_url", "https://osu.ppy.sh")
        self.__create_app_client: bool = kwargs.pop("create_app_client", True)
        self.clients: dict[int, Client] = {}

    async def __aenter__(self) -> ClientStorage:
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        await self.close()

    def on_client_add(self, func: F) -> F:
        """
        A decorator that is called whenever a client is added, to be used as:

            @client_storage.on_client_add

            async def func(event: ClientAddEvent)
        """
        self._register_listener(func, ClientAddEvent)

        @functools.wraps(func)
        async def _on_client_add(*args: Any, **kwargs: Any) -> Any:
            return await func(*args, **kwargs)

        return cast(F, _on_client_add)

    def on_client_update(self, func: F) -> F:
        """
        A decorator that is called whenever any stored client is updated, to be used as:

            @client_storage.on_client_update

            async def func(event: ClientUpdateEvent)
        """
        self._register_listener(func, ClientUpdateEvent)

        @functools.wraps(func)
        async def _on_client_update(*args: Any, **kwargs: Any) -> Any:
            return await func(*args, **kwargs)

        return cast(F, _on_client_update)

    def _get_client_args(self) -> dict[str, Union[str, int]]:
        return {
            "client_secret": self.client_secret,
            "client_id": self.client_id,
            "base_url": self.base_url,
        }

    @property
    async def app_client(self) -> Client:
        r"""Client credentials app client.

        :raises ValueError: If no app client is provided and creation is disabled
        :return: Client credentials app client session
        :rtype: aiosu.v2.client.Client
        """
        if not self.__create_app_client:
            raise ValueError("App clients have been disabled.")

        if 0 not in self.clients:
            client = Client(
                token_repository=self._token_repository,
                session_id=0,
                token=OAuthToken(),
                **self._get_client_args(),
            )
            self.clients[0] = client
            await self._process_event(
                ClientAddEvent(session_id=0, client=client),
            )

        return self.clients[0]

    def client_exists(self, client_uid: int) -> bool:
        r"""Checks if a client exists.

        :param client_uid: The owner user ID of the client
        :type client_uid: int
        :return: Whether the client with the given ID exists
        :rtype: bool
        """
        return client_uid in self.clients

    async def add_client(
        self,
        token: OAuthToken,
        **kwargs: Any,
    ) -> Client:
        r"""Adds a client to storage.

        :param token: The token object of the client
        :type token: aiosu.models.OAuthToken
        :param \**kwargs:
            See below

        :Keyword Arguments:
            * *id* (``int``) --
                Optional, the ID of the client session
        :return: The client with the given ID or token
        :rtype: aiosu.v2.client.Client
        """
        session_id: int = kwargs.pop("id", token.owner_id)
        client = Client(
            token_repository=self._token_repository,
            session_id=session_id,
            token=token,
            **self._get_client_args(),
        )
        client._register_listener(self._process_event, ClientUpdateEvent)
        await client._prepare_token()
        await self._process_event(
            ClientAddEvent(session_id=session_id, client=client),
        )
        self.clients[session_id] = client
        return client

    async def get_client(self, **kwargs: Any) -> Client:
        r"""Gets a client from storage.

        :param \**kwargs:
            See below

        :Keyword Arguments:
            * *id* (``int``) --
                Optional, the owner user ID of the client
            * *token* (``OAuthToken``) --
                Optional, the token object of the client
        :raises InvalidClientRequestedError: If no valid ID or token is provided
        :return: The client with the given ID or token
        :rtype: aiosu.v2.client.Client
        """
        session_id: int = kwargs.pop("id", None)
        token: OAuthToken = kwargs.pop("token", None)
        if self.client_exists(session_id):
            return self.clients[session_id]
        if token is not None:
            return await self.add_client(token, id=session_id)
        if await self._token_repository.exists(session_id):
            token = await self._token_repository.get(session_id)
            return await self.add_client(token, id=session_id)
        raise InvalidClientRequestedError(
            "Either a valid id or token must be specified.",
        )

    async def revoke_client(self, client_uid: int) -> None:
        r"""Revokes a client from storage.

        :param client_uid: The owner user ID of the client
        :type client_uid: int
        :raises InvalidClientRequestedError: If no client exists with the given ID
        """
        if self.client_exists(client_uid):
            await self.clients[client_uid].revoke_token()
            del self.clients[client_uid]
            return

        raise InvalidClientRequestedError(
            "No client exists with the given ID.",
        )

    async def close(self) -> None:
        r"""Closes all client sessions."""
        for client in self.clients.values():
            await client.close()



"""
This module contains models for Beatmap objects.
"""
from __future__ import annotations

from datetime import datetime
from enum import Enum
from enum import unique
from typing import Any
from typing import Literal
from typing import Optional

from pydantic import computed_field
from pydantic import Field
from pydantic import model_validator

from .base import BaseModel
from .common import CurrentUserAttributes
from .common import CursorModel
from .gamemode import Gamemode
from .user import User

__all__ = (
    "Beatmap",
    "BeatmapDescription",
    "BeatmapGenre",
    "BeatmapLanguage",
    "BeatmapAvailability",
    "BeatmapCovers",
    "BeatmapDifficultyAttributes",
    "BeatmapFailtimes",
    "BeatmapHype",
    "BeatmapNominations",
    "BeatmapRankStatus",
    "Beatmapset",
    "BeatmapsetDiscussion",
    "BeatmapsetDiscussionPost",
    "BeatmapsetDisscussionType",
    "BeatmapsetEvent",
    "BeatmapsetEventComment",
    "BeatmapsetEventType",
    "BeatmapsetRequestStatus",
    "BeatmapsetVoteEvent",
    "BeatmapUserPlaycount",
    "BeatmapsetDiscussionResponse",
    "BeatmapsetDiscussionPostResponse",
    "BeatmapsetDiscussionVoteResponse",
    "BeatmapsetSearchResponse",
)

BeatmapsetDisscussionType = Literal[
    "hype",
    "praise",
    "problem",
    "review",
    "suggestion",
    "mapper_note",
]


BeatmapsetEventType = Literal[
    "approve",
    "beatmap_owner_change",
    "discussion_delete",
    "discussion_post_delete",
    "discussion_post_restore",
    "discussion_restore",
    "discussion_lock",
    "discussion_unlock",
    "disqualify",
    "genre_edit",
    "issue_reopen",
    "issue_resolve",
    "kudosu_allow",
    "kudosu_deny",
    "kudosu_gain",
    "kudosu_lost",
    "kudosu_recalculate",
    "language_edit",
    "love",
    "nominate",
    "nomination_reset",
    "nomination_reset_received",
    "nsfw_toggle",
    "offset_edit",
    "qualify",
    "rank",
    "remove_from_loved",
]

BeatmapsetRequestStatus = Literal[
    "all",
    "ranked",
    "qualified",
    "disqualified",
    "never_ranked",
]

BEATMAP_RANK_STATUS_NAMES = {
    -2: "graveyard",
    -1: "wip",
    0: "pending",
    1: "ranked",
    2: "approved",
    3: "qualified",
    4: "loved",
}


@unique
class BeatmapRankStatus(Enum):
    GRAVEYARD = -2
    WIP = -1
    PENDING = 0
    RANKED = 1
    APPROVED = 2
    QUALIFIED = 3
    LOVED = 4

    @property
    def id(self) -> int:
        return self.value

    @property
    def name_api(self) -> str:
        return BEATMAP_RANK_STATUS_NAMES[self.id]

    def __str__(self) -> str:
        return self.name_api

    @classmethod
    def _missing_(cls, query: object) -> Any:
        if isinstance(query, int):
            for status in list(BeatmapRankStatus):
                if status.id == query:
                    return status
        elif isinstance(query, str):
            for status in list(BeatmapRankStatus):
                if status.name_api == query.lower():
                    return status
        raise ValueError(f"BeatmapRankStatus {query} does not exist.")


class BeatmapDescription(BaseModel):
    bbcode: Optional[str] = None
    description: Optional[str] = None


class BeatmapGenre(BaseModel):
    name: str
    id: Optional[int] = None


class BeatmapLanguage(BaseModel):
    name: str
    id: Optional[int] = None


class BeatmapAvailability(BaseModel):
    more_information: Optional[str] = None
    download_disabled: Optional[bool] = None

    @classmethod
    def _from_api_v1(cls, data: Any) -> BeatmapAvailability:
        return cls.model_validate({"download_disabled": data["download_unavailable"]})


class BeatmapNominations(BaseModel):
    current: Optional[int] = None
    required: Optional[int] = None


class BeatmapNomination(BaseModel):
    beatmapset_id: int
    reset: bool
    user_id: int
    rulesets: Optional[list[Gamemode]] = None


class BeatmapCovers(BaseModel):
    cover: str
    card: str
    list: str
    slimcover: str
    cover_2_x: Optional[str] = Field(default=None, alias="cover@2x")
    card_2_x: Optional[str] = Field(default=None, alias="card@2x")
    list_2_x: Optional[str] = Field(default=None, alias="list@2x")
    slimcover_2_x: Optional[str] = Field(default=None, alias="slimcover@2x")

    @classmethod
    def from_beatmapset_id(cls, beatmapset_id: int) -> BeatmapCovers:
        base_url = "https://assets.ppy.sh/beatmaps/"
        return cls.model_validate(
            {
                "cover": f"{base_url}{beatmapset_id}/covers/cover.jpg",
                "card": f"{base_url}{beatmapset_id}/covers/card.jpg",
                "list": f"{base_url}{beatmapset_id}/covers/list.jpg",
                "slimcover": f"{base_url}{beatmapset_id}/covers/slimcover.jpg",
                "cover_2_x": f"{base_url}{beatmapset_id}/covers/cover@2x.jpg",
                "card_2_x": f"{base_url}{beatmapset_id}/covers/card@2x.jpg",
                "list_2_x": f"{base_url}{beatmapset_id}/covers/list@2x.jpg",
                "slimcover_2_x": f"{base_url}{beatmapset_id}/covers/slimcover@2x.jpg",
            },
        )

    @classmethod
    def _from_api_v1(cls, data: Any) -> BeatmapCovers:
        return cls.from_beatmapset_id(data["beatmapset_id"])


class BeatmapHype(BaseModel):
    current: int
    required: int


class BeatmapFailtimes(BaseModel):
    exit: Optional[list[int]] = None
    fail: Optional[list[int]] = None


class BeatmapDifficultyAttributes(BaseModel):
    max_combo: int
    star_rating: float
    # osu standard
    aim_difficulty: Optional[float] = None
    approach_rate: Optional[float] = None  # osu catch + standard
    flashlight_difficulty: Optional[float] = None
    overall_difficulty: Optional[float] = None
    slider_factor: Optional[float] = None
    speed_difficulty: Optional[float] = None
    speed_note_count: Optional[float] = None
    # osu taiko
    stamina_difficulty: Optional[float] = None
    rhythm_difficulty: Optional[float] = None
    colour_difficulty: Optional[float] = None
    # osu mania
    great_hit_window: Optional[float] = None
    score_multiplier: Optional[float] = None


class Beatmap(BaseModel):
    id: int
    url: str
    mode: Gamemode
    beatmapset_id: int
    difficulty_rating: float
    status: BeatmapRankStatus
    total_length: int
    user_id: int
    version: str
    accuracy: Optional[float] = None
    ar: Optional[float] = None
    cs: Optional[float] = None
    bpm: Optional[float] = None
    convert: Optional[bool] = None
    count_circles: Optional[int] = None
    count_sliders: Optional[int] = None
    count_spinners: Optional[int] = None
    deleted_at: Optional[datetime] = None
    drain: Optional[float] = None
    hit_length: Optional[int] = None
    is_scoreable: Optional[bool] = None
    last_updated: Optional[datetime] = None
    passcount: Optional[int] = None
    play_count: Optional[int] = Field(default=None, alias="playcount")
    checksum: Optional[str] = None
    max_combo: Optional[int] = None
    beatmapset: Optional[Beatmapset] = None
    failtimes: Optional[BeatmapFailtimes] = None

    @model_validator(mode="before")
    @classmethod
    def _set_url(cls, values: dict[str, Any]) -> dict[str, Any]:
        if values.get("url") is None:
            id = values["id"]
            beatmapset_id = values["beatmapset_id"]
            mode = Gamemode(values["mode"])
            values[
                "url"
            ] = f"https://osu.ppy.sh/beatmapsets/{beatmapset_id}#{mode}/{id}"
        return values

    @computed_field  # type: ignore
    @property
    def discussion_url(self) -> str:
        return f"https://osu.ppy.sh/beatmapsets/{self.beatmapset_id}/discussion/{self.id}/general"

    @computed_field  # type: ignore
    @property
    def count_objects(self) -> Optional[int]:
        """Total count of the objects.

        :return: Sum of counts of all objects. None if no object count information.
        :rtype: Optional[int]
        """
        if (
            self.count_circles is None
            or self.count_spinners is None
            or self.count_sliders is None
        ):
            return None
        return self.count_spinners + self.count_circles + self.count_sliders

    @classmethod
    def _from_api_v1(cls, data: Any) -> Beatmap:
        return cls.model_validate(
            {
                "beatmapset_id": data["beatmapset_id"],
                "difficulty_rating": data["difficultyrating"],
                "id": data["beatmap_id"],
                "mode": int(data["mode"]),
                "status": int(data["approved"]),
                "total_length": data["total_length"],
                "hit_length": data["total_length"],
                "user_id": data["creator_id"],
                "version": data["version"],
                "accuracy": data["diff_overall"],
                "cs": data["diff_size"],
                "ar": data["diff_approach"],
                "drain": data["diff_drain"],
                "last_updated": data["last_update"],
                "bpm": data["bpm"],
                "checksum": data["file_md5"],
                "playcount": data["playcount"],
                "passcount": data["passcount"],
                "count_circles": data["count_normal"],
                "count_sliders": data["count_slider"],
                "count_spinners": data["count_spinner"],
                "max_combo": data["max_combo"],
            },
        )


class Beatmapset(BaseModel):
    id: int
    artist: str
    artist_unicode: str
    covers: BeatmapCovers
    creator: str
    favourite_count: int
    play_count: int = Field(alias="playcount")
    preview_url: str
    source: str
    status: BeatmapRankStatus
    title: str
    title_unicode: str
    user_id: int
    video: bool
    nsfw: Optional[bool] = None
    hype: Optional[BeatmapHype] = None
    availability: Optional[BeatmapAvailability] = None
    bpm: Optional[float] = None
    can_be_hyped: Optional[bool] = None
    discussion_enabled: Optional[bool] = None
    discussion_locked: Optional[bool] = None
    is_scoreable: Optional[bool] = None
    last_updated: Optional[datetime] = None
    legacy_thread_url: Optional[str] = None
    nominations: Optional[BeatmapNominations] = None
    current_nominations: Optional[list[BeatmapNomination]] = None
    ranked_date: Optional[datetime] = None
    storyboard: Optional[bool] = None
    submitted_date: Optional[datetime] = None
    tags: Optional[str] = None
    pack_tags: Optional[list[str]] = None
    track_id: Optional[int] = None
    related_users: Optional[list[User]] = None
    current_user_attributes: Optional[CurrentUserAttributes] = None
    description: Optional[BeatmapDescription] = None
    genre: Optional[BeatmapGenre] = None
    language: Optional[BeatmapLanguage] = None
    ratings: Optional[list[int]] = None
    has_favourited: Optional[bool] = None
    beatmaps: Optional[list[Beatmap]] = None
    converts: Optional[list[Beatmap]] = None

    @computed_field  # type: ignore
    @property
    def url(self) -> str:
        return f"https://osu.ppy.sh/beatmapsets/{self.id}"

    @computed_field  # type: ignore
    @property
    def discussion_url(self) -> str:
        return f"https://osu.ppy.sh/beatmapsets/{self.id}/discussion"

    @classmethod
    def _from_api_v1(cls, data: Any) -> Beatmapset:
        return cls.model_validate(
            {
                "id": data["beatmapset_id"],
                "artist": data["artist"],
                "artist_unicode": data["artist"],
                "covers": BeatmapCovers._from_api_v1(data),
                "favourite_count": data["favourite_count"],
                "creator": data["creator"],
                "play_count": data["playcount"],
                "preview_url": f"//b.ppy.sh/preview/{data['beatmapset_id']}.mp3",
                "source": data["source"],
                "status": int(data["approved"]),
                "title": data["title"],
                "title_unicode": data["title"],
                "user_id": data["creator_id"],
                "video": data["video"],
                "submitted_date": data["submit_date"],
                "ranked_date": data["approved_date"],
                "last_updated": data["last_update"],
                "tags": data["tags"],
                "storyboard": data["storyboard"],
                "availabiliy": BeatmapAvailability._from_api_v1(data),
                "beatmaps": [Beatmap._from_api_v1(data)],
            },
        )


class BeatmapsetSearchResponse(CursorModel):
    beatmapsets: list[Beatmapset]


class BeatmapUserPlaycount(BaseModel):
    count: int
    beatmap_id: int
    beatmap: Optional[Beatmap] = None
    beatmapset: Optional[Beatmapset] = None


class BeatmapsetDiscussionPost(BaseModel):
    id: int
    user_id: int
    system: bool
    message: str
    created_at: datetime
    beatmap_discussion_id: Optional[int] = None
    last_editor_id: Optional[int] = None
    deleted_by_id: Optional[int] = None
    updated_at: Optional[datetime] = None
    deleted_at: Optional[datetime] = None


class BeatmapsetDiscussion(BaseModel):
    id: int
    beatmapset_id: int
    user_id: int
    message_type: BeatmapsetDisscussionType
    resolved: bool
    can_be_resolved: bool
    can_grant_kudosu: bool
    created_at: datetime
    beatmap_id: Optional[int] = None
    deleted_by_id: Optional[int] = None
    parent_id: Optional[int] = None
    timestamp: Optional[int] = None
    updated_at: Optional[datetime] = None
    deleted_at: Optional[datetime] = None
    last_post_at: Optional[datetime] = None
    kudosu_denied: Optional[bool] = None
    starting_post: Optional[BeatmapsetDiscussionPost] = None


class BeatmapsetVoteEvent(BaseModel):
    score: int
    user_id: int
    id: Optional[int] = None
    created_at: Optional[datetime] = None
    updated_at: Optional[datetime] = None
    beatmapset_discussion_id: Optional[int] = None


class BeatmapsetEventComment(BaseModel):
    beatmap_discussion_id: Optional[int] = None
    beatmap_discussion_post_id: Optional[int] = None
    new_vote: Optional[BeatmapsetVoteEvent] = None
    votes: Optional[list[BeatmapsetVoteEvent]] = None
    mode: Optional[Gamemode] = None
    reason: Optional[str] = None
    source_user_id: Optional[int] = None
    source_user_username: Optional[str] = None
    nominator_ids: Optional[list[int]] = None
    new: Optional[str] = None
    old: Optional[str] = None
    new_user_id: Optional[int] = None
    new_user_username: Optional[str] = None


class BeatmapsetEvent(BaseModel):
    id: int
    type: BeatmapsetEventType
    r"""Information on types: https://github.com/ppy/osu-web/blob/master/resources/assets/lib/interfaces/beatmapset-event-json.ts"""
    created_at: datetime
    user_id: int
    beatmapset: Optional[Beatmapset] = None
    discussion: Optional[BeatmapsetDiscussion] = None
    comment: Optional[dict] = None


class BeatmapsetDiscussionResponse(CursorModel):
    beatmaps: list[Beatmap]
    discussions: list[BeatmapsetDiscussion]
    included_discussions: list[BeatmapsetDiscussion]
    users: list[User]
    max_blocks: int

    @model_validator(mode="before")
    @classmethod
    def _set_max_blocks(cls, values: dict[str, Any]) -> dict[str, Any]:
        values["max_blocks"] = values["reviews_config"]["max_blocks"]
        return values


class BeatmapsetDiscussionPostResponse(CursorModel):
    beatmapsets: list[Beatmapset]
    posts: list[BeatmapsetDiscussionPost]
    users: list[User]


class BeatmapsetDiscussionVoteResponse(CursorModel):
    votes: list[BeatmapsetVoteEvent]
    discussions: list[BeatmapsetDiscussion]
    users: list[User]


Beatmap.model_rebuild()



"""
This module contains miscellaneous helper functions.
"""
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any
    from typing import Callable
    from typing import Optional
    from typing import TypeVar

    T = TypeVar("T")

__all__ = (
    "from_list",
    "add_param",
)


def from_list(f: Callable[[Any], T], x: Any) -> list[T]:
    r"""Applies a function to all elements in a list.

    :param f: Function to apply on list elements
    :type f: Callable[[Any], T]
    :param x: List of objects
    :type x: Any
    :raises TypeError: If x is not a list
    :return: New list
    :rtype: list[T]
    """
    if not isinstance(x, list):
        raise TypeError("Wrong type received. Expected list.")
    return [f(y) for y in x]


def add_param(
    params: dict[str, Any],
    kwargs: dict[str, Any],
    key: str,
    param_name: Optional[str] = None,
    converter: Optional[Callable[[Any], T]] = None,
) -> bool:
    r"""Adds a parameter to a dictionary if it exists in kwargs.

    :param params: Dictionary to add parameter to
    :type params: dict[str, Any]
    :param kwargs: Dictionary to get parameter from
    :type kwargs: dict[str, Any]
    :param key: Key to get parameter from
    :type key: str
    :param param_name: Name of parameter to add to dictionary, defaults to None
    :type param_name: Optional[str]
    :param converter: Function to convert parameter, defaults to None
    :type converter: Optional[Callable[[Any], T]]
    :return: True if parameter was added, False otherwise
    :rtype: bool
    """
    if key in kwargs:
        value = kwargs[key]
        if converter:
            value = converter(value)
        params[param_name or key] = value
        return True
    return False



"""
This module handles API requests for API v2 (OAuth).

You can read more about it here: https://osu.ppy.sh/docs/index.html
"""
from __future__ import annotations

import functools
from datetime import datetime
from functools import partial
from io import BytesIO
from typing import Any
from typing import Callable
from typing import cast
from typing import Literal
from typing import TYPE_CHECKING
from typing import TypeVar
from warnings import warn

import aiohttp
import orjson
from aiolimiter import AsyncLimiter

from ..events import ClientUpdateEvent
from ..events import Eventable
from ..exceptions import APIException
from ..helpers import add_param
from ..helpers import from_list
from ..models import ArtistResponse
from ..models import Beatmap
from ..models import BeatmapDifficultyAttributes
from ..models import Beatmapset
from ..models import BeatmapsetDiscussionPostResponse
from ..models import BeatmapsetDiscussionResponse
from ..models import BeatmapsetDiscussionVoteResponse
from ..models import BeatmapsetEvent
from ..models import BeatmapsetSearchResponse
from ..models import BeatmapUserPlaycount
from ..models import Build
from ..models import ChangelogListing
from ..models import ChatChannel
from ..models import ChatChannelResponse
from ..models import ChatChannelTypes
from ..models import ChatMessage
from ..models import ChatMessageCreateResponse
from ..models import ChatUpdateResponse
from ..models import ChatUserSilence
from ..models import CommentBundle
from ..models import Event
from ..models import ForumCreateTopicResponse
from ..models import ForumPost
from ..models import ForumTopic
from ..models import ForumTopicResponse
from ..models import Gamemode
from ..models import KudosuHistory
from ..models import LazerScore
from ..models import Mods
from ..models import MultiplayerLeaderboardResponse
from ..models import MultiplayerMatchesResponse
from ..models import MultiplayerMatchResponse
from ..models import MultiplayerRoom
from ..models import MultiplayerRoomMode
from ..models import MultiplayerScoresResponse
from ..models import NewsListing
from ..models import NewsPost
from ..models import OAuthToken
from ..models import Rankings
from ..models import RankingType
from ..models import Scopes
from ..models import Score
from ..models import SearchResponse
from ..models import SeasonalBackgroundSet
from ..models import Spotlight
from ..models import User
from ..models import UserQueryType
from ..models import WikiPage
from .repository import BaseTokenRepository
from .repository import SimpleTokenRepository

if TYPE_CHECKING:
    from types import TracebackType
    from typing import Optional
    from typing import Type
    from typing import Union

__all__ = ("Client",)

F = TypeVar("F", bound=Callable[..., Any])
ClientRequestType = Literal["GET", "POST", "DELETE", "PUT", "PATCH"]


def to_lower_str(value: Any) -> str:
    """Converts a value to a lowercase string."""
    return str(value).lower()


def get_content_type(content_type: str) -> str:
    """Returns the content type."""
    return content_type.split(";")[0]


def prepare_token(func: F) -> F:
    """A decorator that prepares the token for use, to be used as:
    @prepare_token
    """

    @functools.wraps(func)
    async def _prepare_token(self: Client, *args: Any, **kwargs: Any) -> Any:
        await self._prepare_token()

        return await func(self, *args, **kwargs)

    return cast(F, _prepare_token)


def check_token(func: F) -> F:
    """
    A decorator that checks the current token, to be used as:
    @check_token
    """

    @functools.wraps(func)
    async def _check_token(self: Client, *args: Any, **kwargs: Any) -> Any:
        token = await self.get_current_token()
        if datetime.utcnow().timestamp() > token.expires_on.timestamp():
            await self._refresh()
        return await func(self, *args, **kwargs)

    return cast(F, _check_token)


def requires_scope(
    required_scopes: Scopes,
    any_scope: bool = False,
) -> Callable[[F], F]:
    """
    A decorator that enforces a scope, to be used as:
    @requires_scope(Scopes.PUBLIC)
    """

    def _requires_scope(
        func: F,
    ) -> F:
        @functools.wraps(func)
        async def _wrap(self: Client, *args: Any, **kwargs: Any) -> Any:
            token = await self.get_current_token()
            if any_scope:
                if not (required_scopes & token.scopes):
                    raise APIException(403, "Missing required scopes.")
            elif required_scopes & token.scopes != required_scopes:
                raise APIException(403, "Missing required scopes.")

            return await func(self, *args, **kwargs)

        return cast(F, _wrap)

    return _requires_scope


class Client(Eventable):
    r"""osu! API v2 Client

    :param \**kwargs:
        See below

    :Keyword Arguments:
        * *token_repository* (``aiosu.v2.repository.BaseTokenRepository``) --
            Optional, defaults to ``aiosu.v2.repository.SimpleTokenRepository()``
        * *session_id* (``int``) --
            Optional, ID of the session to search in the repository, defaults to 0
        * *client_id* (``int``) --
            Optional, required to refresh tokens
        * *client_secret* (``str``) --
            Optional, required to refresh tokens
        * *base_url* (``str``) --
            Optional, base API URL, defaults to "https://osu.ppy.sh"
        * *token* (``aiosu.models.oauthtoken.OAuthToken``) --
            Optional, defaults to client credentials if not provided
        * *limiter* (``tuple[int, int]``) --
            Optional, rate limit, defaults to (600, 60) (600 requests per minute)
    """

    __slots__ = (
        "_token_repository",
        "_initial_token",
        "_session",
        "_limiter",
        "session_id",
        "client_id",
        "client_secret",
        "base_url",
    )

    def __init__(
        self,
        **kwargs: Any,
    ) -> None:
        super().__init__()
        self._register_event(ClientUpdateEvent)
        self._token_repository: BaseTokenRepository = kwargs.pop(
            "token_repository",
            SimpleTokenRepository(),
        )
        max_rate, time_period = kwargs.pop("limiter", (600, 60))
        if (
            not isinstance(self._token_repository, SimpleTokenRepository)
            and "session_id" not in kwargs
        ):
            warn(
                "You are using a custom token repository, but did not provide a session ID. This may cause unexpected behavior.",
            )
        if (max_rate / time_period) > (1000 / 60):
            warn(
                "You are running at an insanely high rate limit. Doing so may result in your account being banned.",
            )
        self.session_id: int = kwargs.pop("session_id", 0)
        self.client_id: int = kwargs.pop("client_id", None)
        self.client_secret: str = kwargs.pop("client_secret", None)
        self._initial_token: Optional[OAuthToken] = kwargs.pop("token", None)
        self.base_url: str = kwargs.pop("base_url", "https://osu.ppy.sh")
        self._limiter: AsyncLimiter = AsyncLimiter(
            max_rate=max_rate,
            time_period=time_period,
        )
        self._session: Optional[aiohttp.ClientSession] = None

    async def __aenter__(self) -> Client:
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        await self.close()

    def on_client_update(
        self,
        func: F,
    ) -> F:
        """
        A decorator that is called whenever a client is updated, to be used as:

            @client.on_client_update

            async def func(event: ClientUpdateEvent)
        """
        self._register_listener(func, ClientUpdateEvent)

        @functools.wraps(func)
        async def _on_client_update(*args: Any, **kwargs: Any) -> Any:
            return await func(*args, **kwargs)

        return cast(F, _on_client_update)

    async def get_current_token(self) -> OAuthToken:
        """Get the current token"""
        return await self._token_repository.get(self.session_id)

    async def _prepare_token(self) -> None:
        """Prepare the token for use."""
        if not await self._token_exists():
            token_to_add = self._initial_token
            if token_to_add is None:
                token_to_add = OAuthToken()
            await self._add_token(token_to_add)
        elif self._initial_token is not None:
            await self._update_token(self._initial_token)
        self._initial_token = None

    async def _add_token(self, token: OAuthToken) -> None:
        """Add a token to the current session"""
        await self._token_repository.add(self.session_id, token)

    async def _update_token(self, token: OAuthToken) -> None:
        """Update the current token"""
        await self._token_repository.update(self.session_id, token)

    async def _token_exists(self) -> bool:
        """Check if a token exists for the current session"""
        return await self._token_repository.exists(self.session_id)

    async def _delete_token(self) -> None:
        """Delete the current token"""
        await self._token_repository.delete(self.session_id)

    async def _get_headers(self) -> dict[str, str]:
        token = await self.get_current_token()
        return {
            "Authorization": f"Bearer {token.access_token}",
            "Content-Type": "application/json",
            "Accept": "application/json",
        }

    async def _refresh_auth_data(self) -> dict[str, Union[str, int]]:
        token = await self.get_current_token()
        return {
            "client_id": self.client_id,
            "client_secret": self.client_secret,
            "grant_type": "refresh_token",
            "refresh_token": token.refresh_token,
        }

    def _refresh_guest_data(self) -> dict[str, Union[str, int]]:
        return {
            "client_id": self.client_id,
            "client_secret": self.client_secret,
            "grant_type": "client_credentials",
            "scope": "public",
        }

    async def _request(
        self,
        request_type: ClientRequestType,
        *args: Any,
        **kwargs: Any,
    ) -> Any:
        await self._prepare_token()

        if self._session is None:
            self._session = aiohttp.ClientSession(headers=await self._get_headers())

        req: dict[str, Callable] = {
            "GET": self._session.get,
            "POST": self._session.post,
            "DELETE": self._session.delete,
            "PUT": self._session.put,
            "PATCH": self._session.patch,
        }

        async with self._limiter:
            async with req[request_type](*args, **kwargs) as resp:
                if resp.status == 204:
                    return

                body = await resp.read()
                content_type = get_content_type(resp.headers.get("content-type", ""))
                if resp.status != 200:
                    json = orjson.loads(body)
                    raise APIException(resp.status, json.get("error", ""))
                if content_type == "application/json":
                    return orjson.loads(body)
                if content_type == "application/octet-stream":
                    return BytesIO(body)
                if content_type.startswith("application/x-osu"):
                    return BytesIO(body)
                if content_type == "text/plain":
                    return body.decode()
                raise APIException(415, f"Unhandled Content Type '{content_type}'")

    async def _refresh(self) -> None:
        r"""INTERNAL: Refreshes the client's token

        :raises APIException: Contains status code and error message
        """
        old_token = await self.get_current_token()
        url = f"{self.base_url}/oauth/token"

        data = {}
        if old_token.can_refresh:
            data = await self._refresh_auth_data()
        else:
            data = self._refresh_guest_data()

        async with aiohttp.ClientSession() as temp_session:
            async with self._limiter:
                async with temp_session.post(url, json=data) as resp:
                    try:
                        body = await resp.read()
                        content_type = get_content_type(
                            resp.headers.get("content-type", ""),
                        )
                        if content_type != "application/json":
                            raise APIException(
                                415,
                                f"Unhandled Content Type '{content_type}'",
                            )
                        json = orjson.loads(body)
                        if resp.status != 200:
                            raise APIException(resp.status, json.get("error", ""))
                        if self._session:
                            await self._session.close()
                        new_token = OAuthToken.model_validate(json)
                        await self._update_token(new_token)
                        self._session = aiohttp.ClientSession(
                            headers=await self._get_headers(),
                        )
                    except aiohttp.client_exceptions.ContentTypeError:
                        raise APIException(403, "Invalid token specified.")

        await self._process_event(
            ClientUpdateEvent(client=self, old_token=old_token, new_token=new_token),
        )

    @prepare_token
    async def get_featured_artists(self, **kwargs: Any) -> ArtistResponse:
        r"""Gets the current featured artists.

        :param \**kwargs:
            See below

        :Keyword Arguments:
            * *limit* (``int``) --
                Optional, the number of featured artists to return.
            * *album* (``str``) --
                Optional, the album to filter by.
            * *artist* (``str``) --
                Optional, the artist to filter by.
            * *genre* (``int``) --
                Optional, the genre ID to filter by.
            * *length* (``list[int]``) --
                Optional, the length range to filter by.
            * *bpm* (``list[int]``) --
                Optional, The BPM range to filter by.
            * *query* (``str``) --
                Optional, the search query to filter by.
            * *is_default_sort* (``bool``) --
                Optional, whether to sort by the default sort.
            * *sort* (``str``) --
                Optional, the sort to use.

        :raises APIException: Contains status code and error message
        :return: Featured artist response object
        :rtype: aiosu.models.artist.ArtistResponse
        """
        url = f"{self.base_url}/beatmaps/artists/tracks"
        params: dict[str, Any] = {}
        add_param(params, kwargs, key="limit")
        add_param(params, kwargs, key="album")
        add_param(params, kwargs, key="artist")
        add_param(params, kwargs, key="genre")
        add_param(params, kwargs, key="length")
        add_param(params, kwargs, key="bpm")
        add_param(params, kwargs, key="query")
        add_param(params, kwargs, key="is_default_sort", converter=to_lower_str)
        add_param(params, kwargs, key="sort")
        add_param(params, kwargs, key="cursor_string")
        json = await self._request("GET", url)
        resp = ArtistResponse.model_validate(json)
        if resp.cursor_string:
            kwargs["cursor_string"] = resp.cursor_string
            resp.next = partial(self.get_featured_artists, **kwargs)
        return resp

    @prepare_token
    async def get_seasonal_backgrounds(self) -> SeasonalBackgroundSet:
        r"""Gets the current seasonal background set.

        :raises APIException: Contains status code and error message
        :return: Seasonal background set object
        :rtype: aiosu.models.backgrounds.SeasonalBackgroundSet
        """
        url = f"{self.base_url}/api/v2/seasonal-backgrounds"
        json = await self._request("GET", url)
        return SeasonalBackgroundSet.model_validate(json)

    @prepare_token
    async def get_changelog_listing(self, **kwargs: Any) -> ChangelogListing:
        r"""Gets the changelog listing.

        :param \**kwargs:
            See below

        :Keyword Arguments:
            * *message_formats* (``list[str]``) --
                Optional, the message formats to return.
            * *from* (``str``) --
                Optional, the start date to return.
            * *to* (``str``) --
                Optional, the end date to return.
            * *max_id* (``int``) --
                Optional, the maximum ID to return.
            * *stream* (``str``) --
                Optional, the stream to return.
            * *cursor_string* (``str``) --
                Optional, the cursor string to use.

        :raises APIException: Contains status code and error message
        :return: Changelog listing object
        :rtype: aiosu.models.changelog.ChangelogListing
        """
        url = f"{self.base_url}/api/v2/changelog"
        params: dict[str, Any] = {
            "message_formats": kwargs.pop("message_formats", ["html", "markdown"]),
        }
        add_param(params, kwargs, key="from")
        add_param(params, kwargs, key="to")
        add_param(params, kwargs, key="max_id")
        add_param(params, kwargs, key="stream")
        add_param(params, kwargs, key="cursor_string")
        json = await self._request("GET", url, params=params)
        resp = ChangelogListing.model_validate(json)
        if resp.cursor_string:  # Unused: API does not return cursor_string
            kwargs["cursor_string"] = resp.cursor_string
            resp.next = partial(self.get_changelog_listing, **kwargs)
        return resp

    @prepare_token
    async def get_changelog_build(self, stream: str, build: str) -> Build:
        r"""Gets a specific build from the changelog.

        :param stream: The stream to get the build from
        :param build: The build to get
        :raises APIException: Contains status code and error message
        :return: Build object
        :rtype: aiosu.models.changelog.Build
        """
        url = f"{self.base_url}/api/v2/changelog/{stream}/{build}"
        json = await self._request("GET", url)
        return Build.model_validate(json)

    @prepare_token
    async def lookup_changelog_build(
        self,
        changelog_query: Union[str, int],
        **kwargs: Any,
    ) -> Build:
        r"""Looks up a build from the changelog.

        :param changelog_query: The query to search for
        :type changelog_query: Union[str, int]
        :param \**kwargs:
            See below

        :Keyword Arguments:
            * *is_id* (``bool``) --
                Optional, whether the query is an ID or not, defaults to ``True`` if the query is an int
            * *message_formats* (``list[aiosu.models.news.ChangelogMessageFormats]``) --
                Optional, message formats to get, defaults to ``["html", "markdown"]``

        :raises APIException: Contains status code and error message
        :return: Build object
        :rtype: aiosu.models.changelog.Build
        """
        url = f"{self.base_url}/api/v2/changelog/{changelog_query}"
        params: dict[str, Any] = {
            "message_formats": kwargs.pop("message_formats", ["html", "markdown"]),
        }
        if "is_id" in kwargs or isinstance(changelog_query, int):
            params["key"] = "id"
        json = await self._request("GET", url, params=params)
        return Build.model_validate(json)

    @prepare_token
    async def get_news_listing(self, **kwargs: Any) -> NewsListing:
        r"""Gets the news listing.

        :param \**kwargs:
            See below

        :Keyword Arguments:
            * *limit* (``int``) --
                Optional, the number of news posts to return. Min: 1, Max: 21, defaults to 12
            * *year* (``int``) --
                Optional, the year to filter by.
            * *cursor_string* (``str``) --
                Optional, the cursor string to use for pagination.

        :raises APIException: Contains status code and error message
        :return: News listing object
        :rtype: aiosu.models.news.NewsListing
        """
        url = f"{self.base_url}/api/v2/news"
        if not 1 <= (limit := kwargs.pop("limit", 12)) <= 21:
            raise ValueError("Invalid limit specified. Limit must be between 1 and 21")
        params: dict[str, Any] = {
            "limit": limit,
        }
        add_param(params, kwargs, key="year")
        add_param(params, kwargs, key="cursor_string")
        json = await self._request("GET", url, params=params)
        resp = NewsListing.model_validate(json)
        if resp.cursor_string:
            kwargs["cursor_string"] = resp.cursor_string
            resp.next = partial(self.get_news_listing, **kwargs)
        return resp

    @prepare_token
    async def get_news_post(
        self,
        news_query: Union[str, int],
        **kwargs: Any,
    ) -> NewsPost:
        r"""Gets a news post.

        :param news_query: The query to search for
        :type news_query: Union[str, int]
        :param \**kwargs:
            See below

        :Keyword Arguments:
            * *is_id* (``bool``) --
                Optional, whether the query is an ID or not, defaults to ``True`` if the query is an int

        :raises APIException: Contains status code and error message
        :return: News post object
        :rtype: aiosu.models.news.NewsPost
        """
        url = f"{self.base_url}/api/v2/news/{news_query}"
        params: dict[str, Any] = {
            "message_formats": kwargs.pop("message_formats", ["html", "markdown"]),
        }
        if "is_id" in kwargs or isinstance(news_query, int):
            params["key"] = "id"
        json = await self._request("GET", url, params=params)
        return NewsPost.model_validate(json)

    @prepare_token
    async def get_wiki_page(self, locale: str, path: str) -> WikiPage:
        r"""Gets a wiki page.

        :param locale: The locale of the wiki page
        :type locale: str
        :param path: The path of the wiki page
        :type path: str
        :raises APIException: Contains status code and error message
        :return: Wiki page object
        :rtype: aiosu.models.wiki.WikiPage
        """
        url = f"{self.base_url}/api/v2/wiki/{locale}/{path}"
        json = await self._request("GET", url)
        return WikiPage.model_validate(json)

    @prepare_token
    async def get_comment(self, comment_id: int, **kwargs: Any) -> CommentBundle:
        r"""Gets a comment.

        :param comment_id: The ID of the comment
        :type comment_id: int
        :param \**kwargs:
            See below

        :Keyword Arguments:
            * *cursor_string* (``str``) --
                Optional, cursor string to get the next page of comments

        :raises APIException: Contains status code and error message
        :return: Comment bundle object
        :rtype: aiosu.models.comment.CommentBundle
        """
        url = f"{self.base_url}/api/v2/comments/{comment_id}"
        params: dict[str, Any] = {}
        add_param(params, kwargs, key="cursor_string")
        json = await self._request("GET", url, params=params)
        resp = CommentBundle.model_validate(json)
        if resp.cursor_string:  # Unused: API does not return cursor_string
            kwargs["cursor_string"] = resp.cursor_string
            resp.next = partial(self.get_comment, comment_id=comment_id, **kwargs)
        return resp

    @prepare_token
    async def get_comments(self, **kwargs: Any) -> CommentBundle:
        r"""Gets comments.

        :param \**kwargs:
            See below

        :Keyword Arguments:
            * *commentable_type* (``Literal["beatmapset", "build", "news_post", "user"]``) --
                Optional, commentable type to get comments from
            * *commentable_id* (``int``) --
                Optional, commentable ID to get comments from
            * *parent_id* (``int``) --
                Optional, parent ID to get comments from
            * *sort* (aiosu.models.comment.CommentSortType) --
                Optional, sort order of comments, defaults to ``"new"``
            * *cursor_string* (``str``) --
                Optional, cursor string to get the next page of comments

        :raises APIException: Contains status code and error message
        :return: Comment bundle object
        :rtype: aiosu.models.comment.CommentBundle
        """
        url = f"{self.base_url}/api/v2/comments"
        params: dict[str, Any] = {}
        add_param(params, kwargs, key="commentable_type")
        add_param(params, kwargs, key="commentable_id")
        add_param(params, kwargs, key="parent_id")
        add_param(params, kwargs, key="sort")
        add_param(params, kwargs, key="cursor_string")
        json = await self._request("GET", url, params=params)
        resp = CommentBundle.model_validate(json)
        if resp.cursor_string:  # Unused: API does not return cursor_string
            kwargs["cursor_string"] = resp.cursor_string
            resp.next = partial(self.get_comments, **kwargs)
        return resp

    @prepare_token
    @check_token
    @requires_scope(Scopes.PUBLIC)
    async def search(self, query: str, **kwargs: Any) -> SearchResponse:
        r"""Searches for a user, beatmap, beatmapset, or wiki page.

        :param query: The query to search for
        :type query: str
        :param \**kwargs:
            See below

        :Keyword Arguments:
            * *mode* (``Literal["all", "user", "wiki_page"]``) --
                Optional, gamemode to search for, defaults to ``all``
            * *page* (``int``) --
                Optional, page to get, ignored if mode is ``all``

        :raises APIException: Contains status code and error message
        :return: Search response object
        :rtype: aiosu.models.search.SearchResponse
        """
        url = f"{self.base_url}/api/v2/search"
        params: dict[str, Any] = {
            "query": query,
            "mode": kwargs.pop("mode", "all"),
        }
        add_param(params, kwargs, key="page")
        json = await self._request("GET", url, params=params)
        return SearchResponse.model_validate(json)

    @prepare_token
    @check_token
    @requires_scope(Scopes.IDENTIFY | Scopes.DELEGATE, any_scope=True)
    async def get_me(self, **kwargs: Any) -> User:
        r"""Gets the user who owns the current token

        :param \**kwargs:
            See below

        :Keyword Arguments:
            * *mode* (``aiosu.models.gamemode.Gamemode``) --
                Optional, gamemode to search for

        :raises APIException: Contains status code and error message
        :return: Requested user
        :rtype: aiosu.models.user.User
        """
        url = f"{self.base_url}/api/v2/me"
        if "mode" in kwargs:
            mode = Gamemode(kwargs.pop("mode"))
            url += f"/{mode}"
        json = await self._request("GET", url)
        return User.model_validate(json)

    @prepare_token
    @check_token
    @requires_scope(Scopes.FRIENDS_READ)
    @requires_scope(Scopes.IDENTIFY | Scopes.DELEGATE, any_scope=True)
    async def get_own_friends(self) -> list[User]:
        r"""Gets the token owner's friend list

        :raises APIException: Contains status code and error message
        :return: List of friends
        :rtype: list[aiosu.models.user.User]
        """
        url = f"{self.base_url}/api/v2/friends"
        json = await self._request("GET", url)
        return from_list(User.model_validate, json)

    @prepare_token
    @check_token
    @requires_scope(Scopes.PUBLIC)
    async def get_user(self, user_query: Union[str, int], **kwargs: Any) -> User:
        r"""Gets a user by a query.

        :param user_query: Username or ID to search by
        :type user_query: Union[str, int]
        :param \**kwargs:
            See below

        :Keyword Arguments:
            * *mode* (``aiosu.models.gamemode.Gamemode``) --
                Optional, gamemode to search for
            * *qtype* (``str``) --
                Optional, "string" or "id". Type of the user_query

        :raises APIException: Contains status code and error message
        :return: Requested user
        :rtype: aiosu.models.user.User
        """
        url = f"{self.base_url}/api/v2/users/{user_query}"
        params: dict[str, Any] = {}
        if "mode" in kwargs:
            mode = Gamemode(kwargs.pop("mode"))
            url += f"/{mode}"
        add_param(
            params,
            kwargs,
            key="qtype",
            param_name="type",
            converter=lambda x: UserQueryType(x).new_api_name,
        )
        json = await self._request("GET", url, params=params)
        return User.model_validate(json)

    @prepare_token
    @check_token
    @requires_scope(Scopes.PUBLIC)
    async def get_users(self, user_ids: list[int]) -> list[User]:
        r"""Get multiple user data.

        :param user_ids: The IDs of the users
        :type user_ids: list[int]
        :raises APIException: Contains status code and error message
        :return: List of user data objects
        :rtype: list[aiosu.models.user.User]
        """
        url = f"{self.base_url}/api/v2/users"
        params: dict[str, Any] = {
            "ids": user_ids,
        }
        json = await self._request("GET", url, params=params)
        return from_list(User.model_validate, json.get("users", []))

    @prepare_token
    @check_token
    @requires_scope(Scopes.PUBLIC)
    async def get_user_kudosu(self, user_id: int, **kwargs: Any) -> list[KudosuHistory]:
        r"""Get a user's kudosu history.

        :param user_id: User ID to search by
        :type user_id: int
        :param \**kwargs:
            See below

        :Keyword Arguments:
            * *limit* (``int``) --
                Optional, number of scores to get
            * *offset* (``int``) --
                Optional, offset of the first score to get

        :raises APIException: Contains status code and error message
        :return: List of kudosu history objects
        :rtype: list[aiosu.models.kudosu.KudosuHistory]
        """
        url = f"{self.base_url}/api/v2/users/{user_id}/kudosu"
        params: dict[str, Any] = {}
        add_param(params, kwargs, key="limit")
        add_param(params, kwargs, key="offset")
        json = await self._request("GET", url, params=params)
        return from_list(KudosuHistory.model_validate, json)

    @prepare_token
    @check_token
    @requires_scope(Scopes.PUBLIC)
    async def __get_type_scores(
        self,
        user_id: int,
        request_type: str,
        **kwargs: Any,
    ) -> list[Union[Score, LazerScore]]:
        r"""INTERNAL: Get a user's scores by type

        :param user_id: User ID to search by
        :type user_id: int
        :param request_type: "recent", "best" or "firsts"
        :type request_type: str
        :param \**kwargs:
            See below

        :Keyword Arguments:
            * *limit* (``int``) --
                Optional, number of scores to get. Min: 1, Max: 100, defaults to 100
            * *offset* (``int``) --
                Optional, page offset to start from, defaults to 0
            * *mode* (``aiosu.models.gamemode.Gamemode``) --
                Optional, gamemode to search for
            * *include_fails* (``bool``) --
                Optional, whether to include failed scores, defaults to ``False``
            * *new_format* (``bool``) --
                Optional, whether to use the new format, defaults to ``False``

        :raises ValueError: If limit is not between 1 and 100
        :raises ValueError: If type is invalid
        :raises APIException: Contains status code and error message
        :return: List of requested scores
        :rtype: list[aiosu.models.score.Score] or list[aiosu.models.score.LazerScore]
        """
        if not 1 <= (limit := kwargs.pop("limit", 100)) <= 100:
            raise ValueError("Invalid limit specified. Limit must be between 1 and 100")
        if request_type not in ("recent", "best", "firsts", "pinned"):
            raise ValueError(
                f'"{request_type}" is not a valid request_type. Valid options are: "recent", "best", "firsts"',
            )
        url = f"{self.base_url}/api/v2/users/{user_id}/scores/{request_type}"
        params: dict[str, Any] = {
            "include_fails": int(kwargs.pop("include_fails", False)),
            "limit": limit,
            "offset": kwargs.pop("offset", 0),
        }
        add_param(params, kwargs, key="mode", converter=lambda x: str(Gamemode(x)))
        headers = {}
        new_format = kwargs.pop("new_format", False)
        if new_format:
            headers = {"x-api-version": "20220705"}
        json = await self._request("GET", url, params=params, headers=headers)
        if new_format:
            return from_list(LazerScore.model_validate, json)
        return from_list(Score.model_validate, json)

    async def get_user_recents(
        self,
        user_id: int,
        **kwargs: Any,
    ) -> list[Union[Score, LazerScore]]:
        r"""Get a user's recent scores.

        :param user_id: User ID to search by
        :type user_id: int
        :param \**kwargs:
            See below

        :Keyword Arguments:
            * *mode* (``aiosu.models.gamemode.Gamemode``) --
                Optional, gamemode to search for
            * *limit* (``int``) --
                Optional, number of scores to get. Min: 1, Max: 100, defaults to 100
            * *include_fails* (``bool``) --
                Optional, whether to include failed scores, defaults to ``False``
            * *offset* (``int``) --
                Optional, page offset to start from, defaults to 0
            * *new_format* (``bool``) --
                Optional, whether to use the new format, defaults to ``False``

        :raises APIException: Contains status code and error message
        :return: List of requested scores
        :rtype: list[aiosu.models.score.Score] or list[aiosu.models.score.LazerScore]
        """
        return await self.__get_type_scores(user_id, "recent", **kwargs)

    async def get_user_bests(
        self,
        user_id: int,
        **kwargs: Any,
    ) -> list[Union[Score, LazerScore]]:
        r"""Get a user's top scores.

        :param user_id: User ID to search by
        :type user_id: int
        :param \**kwargs:
            See below

        :Keyword Arguments:
            * *mode* (``aiosu.models.gamemode.Gamemode``) --
                Optional, gamemode to search for
            * *limit* (``int``) --
                Optional, number of scores to get. Min: 1, Max: 100, defaults to 100
            * *offset* (``int``) --
                Optional, page offset to start from, defaults to 0
            * *new_format* (``bool``) --
                Optional, whether to use the new format, defaults to ``False``

        :raises APIException: Contains status code and error message
        :return: List of requested scores
        :rtype: list[aiosu.models.score.Score] or list[aiosu.models.score.LazerScore]
        """
        return await self.__get_type_scores(user_id, "best", **kwargs)

    async def get_user_firsts(
        self,
        user_id: int,
        **kwargs: Any,
    ) -> list[Union[Score, LazerScore]]:
        r"""Get a user's first place scores.

        :param user_id: User ID to search by
        :type user_id: int
        :param \**kwargs:
            See below

        :Keyword Arguments:
            * *mode* (``aiosu.models.gamemode.Gamemode``) --
                Optional, gamemode to search for
            * *limit* (``int``) --
                Optional, number of scores to get. Min: 1, Max: 100, defaults to 100
            * *offset* (``int``) --
                Optional, page offset to start from, defaults to 0
            * *new_format* (``bool``) --
                Optional, whether to use the new format, defaults to ``False``

        :raises APIException: Contains status code and error message
        :return: List of requested scores
        :rtype: list[aiosu.models.score.Score] or list[aiosu.models.score.LazerScore]
        """
        return await self.__get_type_scores(user_id, "firsts", **kwargs)

    async def get_user_pinned(
        self,
        user_id: int,
        **kwargs: Any,
    ) -> list[Union[Score, LazerScore]]:
        r"""Get a user's pinned scores.

        :param user_id: User ID to search by
        :type user_id: int
        :param \**kwargs:
            See below

        :Keyword Arguments:
            * *mode* (``aiosu.models.gamemode.Gamemode``) --
                Optional, gamemode to search for
            * *limit* (``int``) --
                Optional, number of scores to get. Min: 1, Max: 100, defaults to 100
            * *offset* (``int``) --
                Optional, page offset to start from, defaults to 0
            * *new_format* (``bool``) --
                Optional, whether to use the new format, defaults to ``False``

        :raises APIException: Contains status code and error message
        :return: List of requested scores
        :rtype: list[aiosu.models.score.Score] or list[aiosu.models.score.LazerScore]
        """
        return await self.__get_type_scores(user_id, "pinned", **kwargs)

    @prepare_token
    @check_token
    @requires_scope(Scopes.PUBLIC)
    async def get_user_beatmap_scores(
        self,
        user_id: int,
        beatmap_id: int,
        **kwargs: Any,
    ) -> list[Score]:
        r"""Get a user's scores on a specific beatmap.

        :param user_id: User ID to search by
        :type user_id: int
        :param beatmap_id: Beatmap ID to search by
        :type beatmap_id: int
        :param \**kwargs:
            See below

        :Keyword Arguments:
            * *mode* (``aiosu.models.gamemode.Gamemode``) --
                Optional, gamemode to search for

        :raises APIException: Contains status code and error message
        :return: List of requested scores
        :rtype: list[aiosu.models.score.Score]
        """
        url = f"{self.base_url}/api/v2/beatmaps/{beatmap_id}/scores/users/{user_id}/all"
        params: dict[str, Any] = {}
        add_param(params, kwargs, key="mode", converter=lambda x: str(Gamemode(x)))
        json = await self._request("GET", url, params=params)
        return from_list(Score.model_validate, json.get("scores", []))

    UserBeatmapType = Literal["favourite", "graveyard", "loved", "ranked", "pending"]

    @prepare_token
    @check_token
    @requires_scope(Scopes.PUBLIC)
    async def get_user_beatmaps(
        self,
        user_id: int,
        type: UserBeatmapType,
        **kwargs: Any,
    ) -> list[Beatmapset]:
        r"""Get a user's beatmaps.

        :param user_id: ID of the user
        :type user_id: int
        :param type: Type of beatmaps to get
        :type type: UserBeatmapType
        :param \**kwargs:
            See below

        :Keyword Arguments:
            * *limit* (``int``) --
                Optional, number of beatmaps to get
            * *offset* (``int``) --
                Optional, offset of the first beatmap to get

        :raises APIException: Contains status code and error message
        :return: List of requested beatmaps
        :rtype: list[aiosu.models.beatmap.Beatmap]
        """
        url = f"{self.base_url}/api/v2/users/{user_id}/beatmapsets/{type}"
        params: dict[str, Any] = {}
        add_param(params, kwargs, key="limit")
        add_param(params, kwargs, key="offset")
        json = await self._request("GET", url, params=params)
        return from_list(Beatmapset.model_validate, json)

    @prepare_token
    @check_token
    @requires_scope(Scopes.PUBLIC)
    async def get_user_most_played(
        self,
        user_id: int,
        **kwargs: Any,
    ) -> list[BeatmapUserPlaycount]:
        r"""Get a user's most played beatmaps.

        :param user_id: ID of the user
        :type user_id: int
        :param \**kwargs:
            See below

        :Keyword Arguments:
            * *limit* (``int``) --
                Optional, number of beatmaps to get
            * *offset* (``int``) --
                Optional, offset of the first beatmap to get

        :raises APIException: Contains status code and error message
        :return: List of user playcount objects
        :rtype: list[aiosu.models.beatmap.BeatmapUserPlaycount]
        """
        url = f"{self.base_url}/api/v2/users/{user_id}/beatmapsets/most_played"
        params: dict[str, Any] = {}
        add_param(params, kwargs, key="limit")
        add_param(params, kwargs, key="offset")
        json = await self._request("GET", url, params=params)
        return from_list(BeatmapUserPlaycount.model_validate, json)

    @prepare_token
    @check_token
    @requires_scope(Scopes.PUBLIC)
    async def get_user_recent_activity(
        self,
        user_id: int,
        **kwargs: Any,
    ) -> list[Event]:
        r"""Get a user's recent activity.

        :param user_id: ID of the user
        :type user_id: int
        :param \**kwargs:
            See below

        :Keyword Arguments:
            * *limit* (``int``) --
                Optional, number of events to get
            * *offset* (``int``) --
                Optional, offset of the first event to get

        :raises APIException: Contains status code and error message
        :return: List of events
        :rtype: list[aiosu.models.event.Event]
        """
        url = f"{self.base_url}/api/v2/users/{user_id}/recent_activity"
        params: dict[str, Any] = {}
        add_param(params, kwargs, key="limit")
        add_param(params, kwargs, key="offset")
        json = await self._request("GET", url, params=params)
        return from_list(Event.model_validate, json)

    @prepare_token
    @check_token
    @requires_scope(Scopes.PUBLIC)
    async def get_beatmap_scores(self, beatmap_id: int, **kwargs: Any) -> list[Score]:
        r"""Get scores submitted on a specific beatmap.

        :param beatmap_id: Beatmap ID to search by
        :type beatmap_id: int
        :param \**kwargs:
            See below

        :Keyword Arguments:
            * *mode* (``aiosu.models.gamemode.Gamemode``) --
                Optional, gamemode to search for
            * *mods* (``aiosu.models.mods.Mods``) --
                Optional, mods to search for
            * *type* (``str``) --
                Optional, beatmap score ranking type

        :raises APIException: Contains status code and error message
        :return: List of requested scores
        :rtype: list[aiosu.models.score.Score]
        """
        url = f"{self.base_url}/api/v2/beatmaps/{beatmap_id}/scores"
        params: dict[str, Any] = {}
        add_param(params, kwargs, key="mode", converter=lambda x: str(Gamemode(x)))
        add_param(params, kwargs, key="mods", converter=lambda x: str(Mods(x)))
        add_param(params, kwargs, key="type")
        json = await self._request("GET", url, params=params)
        return from_list(Score.model_validate, json.get("scores", []))

    @prepare_token
    @check_token
    @requires_scope(Scopes.PUBLIC)
    async def get_beatmap(self, beatmap_id: int) -> Beatmap:
        r"""Get beatmap data.

        :param beatmap_id: The ID of the beatmap
        :type beatmap_id: int
        :raises APIException: Contains status code and error message
        :return: Beatmap data object
        :rtype: aiosu.models.beatmap.Beatmap
        """
        url = f"{self.base_url}/api/v2/beatmaps/{beatmap_id}"
        json = await self._request("GET", url)
        return Beatmap.model_validate(json)

    @prepare_token
    @check_token
    @requires_scope(Scopes.PUBLIC)
    async def get_beatmaps(self, beatmap_ids: list[int]) -> list[Beatmap]:
        r"""Get multiple beatmap data.

        :param beatmap_ids: The IDs of the beatmaps
        :type beatmap_ids: list[int]
        :raises APIException: Contains status code and error message
        :return: List of beatmap data objects
        :rtype: list[aiosu.models.beatmap.Beatmap]
        """
        url = f"{self.base_url}/api/v2/beatmaps"
        params: dict[str, Any] = {
            "ids": beatmap_ids,
        }
        json = await self._request("GET", url, params=params)
        return from_list(Beatmap.model_validate, json.get("beatmaps", []))

    @prepare_token
    @check_token
    @requires_scope(Scopes.PUBLIC)
    async def lookup_beatmap(self, **kwargs: Any) -> Beatmap:
        r"""Lookup beatmap data.

        :param \**kwargs:
            See below

        :Keyword Arguments:
            * *checksum* (``str``) --
                Optional, the MD5 checksum of the beatmap
            * *filename* (``str``) --
                Optional, the filename of the beatmap
            * *id* (``int``) --
                Optional, the ID of the beatmap

        :raises ValueError: If no arguments are specified
        :raises APIException: Contains status code and error message
        :return: Beatmap data object
        :rtype: aiosu.models.beatmap.Beatmap
        """
        url = f"{self.base_url}/api/v2/beatmaps/lookup"
        params: dict[str, Any] = {}
        add_param(params, kwargs, key="checksum")
        add_param(params, kwargs, key="filename")
        add_param(params, kwargs, key="id")
        if not params:
            raise ValueError("One of checksum, filename or id must be provided.")
        json = await self._request("GET", url, params=params)
        return Beatmap.model_validate(json)

    @prepare_token
    @check_token
    @requires_scope(Scopes.PUBLIC)
    async def get_beatmap_attributes(
        self,
        beatmap_id: int,
        **kwargs: Any,
    ) -> BeatmapDifficultyAttributes:
        r"""Gets difficulty attributes for a beatmap.

        :param beatmap_id: The ID of the beatmap
        :type beatmap_id: int
        :param \**kwargs:
            See below

        :Keyword Arguments:
            * *mode* (``aiosu.models.gamemode.Gamemode``) --
                Optional, gamemode to search for
            * *mods* (``aiosu.models.mods.Mods``) --
                Optional, mods to apply to the result

        :raises APIException: Contains status code and error message
        :return: Difficulty attributes for a beatmap
        :rtype: aiosu.models.beatmap.BeatmapDifficultyAttributes
        """
        url = f"{self.base_url}/api/v2/beatmaps/{beatmap_id}/attributes"
        data: dict[str, Any] = {}
        add_param(
            data,
            kwargs,
            key="mode",
            param_name="ruleset_id",
            converter=lambda x: int(Gamemode(x)),
        )
        add_param(data, kwargs, key="mods", converter=lambda x: int(Mods(x)))
        json = await self._request("POST", url, json=data)
        return BeatmapDifficultyAttributes.model_validate(json.get("attributes"))

    @prepare_token
    @check_token
    @requires_scope(Scopes.PUBLIC)
    async def get_beatmapset(self, beatmapset_id: int) -> Beatmapset:
        r"""Get beatmapset data.

        :param beatmapset_id: The ID of the beatmapset
        :type beatmapset_id: int
        :raises APIException: Contains status code and error message
        :return: Beatmapset data object
        :rtype: aiosu.models.beatmap.Beatmapset
        """
        url = f"{self.base_url}/api/v2/beatmapsets/{beatmapset_id}"
        json = await self._request("GET", url)
        return Beatmapset.model_validate(json)

    @prepare_token
    @check_token
    @requires_scope(Scopes.PUBLIC)
    async def lookup_beatmapset(self, beatmap_id: int) -> Beatmapset:
        r"""Lookup beatmap data.

        :param beatmap_id: The ID of a beatmap in the set
        :type beatmap_id: int

        :raises APIException: Contains status code and error message
        :return: Beatmapset data object
        :rtype: aiosu.models.beatmap.Beatmapset
        """
        url = f"{self.base_url}/api/v2/beatmapsets/lookup"
        params: dict[str, Any] = {
            "beatmap_id": beatmap_id,
        }
        json = await self._request("GET", url, params=params)
        return Beatmapset.model_validate(json)

    @prepare_token
    @check_token
    @requires_scope(Scopes.PUBLIC)
    async def search_beatmapsets(
        self,
        search_filter: Optional[str] = "",
        **kwargs: Any,
    ) -> BeatmapsetSearchResponse:
        r"""Search beatmapset by filter.

        :param search_filter: The search filter.
        :type search_filter: str
        :param \**kwargs:
            See below

        :Keyword Arguments:
            * *cursor_string* (``str``) --
                Optional, cursor string to get the next page of results

        :raises APIException: Contains status code and error message
        :return: Beatmapset search response
        :rtype: list[aiosu.models.beatmap.BeatmapsetSearchResponse]
        """
        url = f"{self.base_url}/api/v2/beatmapsets/search/{search_filter}"
        params: dict[str, Any] = {}
        add_param(params, kwargs, key="cursor_string")
        json = await self._request("GET", url)
        resp = BeatmapsetSearchResponse.model_validate(json)
        if resp.cursor_string:
            kwargs["cursor_string"] = resp.cursor_string
            resp.next = partial(self.search_beatmapsets, **kwargs)
        return resp

    @prepare_token
    @check_token
    @requires_scope(Scopes.PUBLIC)
    async def get_beatmapset_events(self, **kwargs: Any) -> list[BeatmapsetEvent]:
        r"""Get beatmapset events.

        :param \**kwargs:
            See below

        :Keyword Arguments:
            * *limit* (``int``) --
                Optional, number of results per page
            * *page* (``int``) --
                Optional, page number
            * *user_id* (``int``) --
                Optional, user ID
            * *min_date* (``datetime.datetime``) --
                Optional, minimum date
            * *max_date* (``datetime.datetime``) --
                Optional, maximum date
            * *types* (``list[aiosu.models.beatmap.BeatmapsetEventType]``) --
                Optional, event types

        :raises APIException: Contains status code and error message
        :return: List of beatmapset events
        :rtype: list[aiosu.models.event.Event]
        """
        url = f"{self.base_url}/api/v2/beatmapsets/events"
        params: dict[str, Any] = {}
        add_param(params, kwargs, key="limit")
        add_param(params, kwargs, key="page")
        add_param(params, kwargs, key="user_id", param_name="user")
        add_param(params, kwargs, key="min_date")
        add_param(params, kwargs, key="max_date")
        add_param(params, kwargs, key="types")
        json = await self._request("GET", url, params=params)
        return from_list(BeatmapsetEvent.model_validate, json.get("events", []))

    @prepare_token
    @check_token
    @requires_scope(Scopes.PUBLIC)
    async def get_beatmapset_discussions(
        self,
        **kwargs: Any,
    ) -> BeatmapsetDiscussionResponse:
        r"""Get beatmapset discussions.

        :param \**kwargs:
            See below

        :Keyword Arguments:
            * *beatmap_id* (``int``) --
                Optional, beatmap ID
            * *beatmapset_id* (``int``) --
                Optional, beatmapset ID
            * *beatmapset_status* (``aiosu.models.beatmap.BeatmapsetRequestStatus``) --
                Optional, beatmapset status
            * *limit* (``int``) --
                Optional, number of results per page
            * *page* (``int``) --
                Optional, page number
            * *message_types* (``list[aiosu.models.beatmap.BeatmapsetDisscussionType]``) --
                Optional, message types
            * *only_unresolved* (``bool``) --
                Optional, only unresolved discussions
            * *sort* (``aiosu.models.common.SortTypes``) --
                Optional, sort order, defaults to ``id_desc``
            * *user_id* (``int``) --
                Optional, user ID
            * with_deleted (``bool``) --
                Optional, include deleted discussions
            * cursor_string (``str``) --
                Optional, cursor string

        :raises APIException: Contains status code and error message
        :return: Beatmapset discussion response
        :rtype: aiosu.models.beatmap.BeatmapsetDiscussionResponse
        """
        url = f"{self.base_url}/api/v2/beatmapsets/discussions"
        params: dict[str, Any] = {}
        add_param(params, kwargs, key="beatmap_id")
        add_param(params, kwargs, key="beatmapset_id")
        add_param(params, kwargs, key="beatmapset_status")
        add_param(params, kwargs, key="limit")
        add_param(params, kwargs, key="page")
        add_param(params, kwargs, key="message_types")
        add_param(params, kwargs, key="only_unresolved", converter=to_lower_str)
        add_param(params, kwargs, key="sort")
        add_param(params, kwargs, key="user", param_name="user_id")
        add_param(params, kwargs, key="with_deleted", converter=to_lower_str)
        add_param(params, kwargs, key="cursor_string")
        json = await self._request("GET", url, params=params)
        resp = BeatmapsetDiscussionResponse.model_validate(json)
        if resp.cursor_string:
            kwargs["cursor_string"] = resp.cursor_string
            resp.next = partial(self.get_beatmapset_discussions, **kwargs)
        return resp

    @prepare_token
    @check_token
    @requires_scope(Scopes.PUBLIC)
    async def get_beatmapset_discussion_posts(
        self,
        **kwargs: Any,
    ) -> BeatmapsetDiscussionPostResponse:
        r"""Get beatmapset discussion posts.

        :param \**kwargs:
            See below

        :Keyword Arguments:
            * *beatmapset_discussion_id* (``int``) --
                Optional, beatmapset discussion ID
            * *limit* (``int``) --
                Optional, number of results per page
            * *page* (``int``) --
                Optional, page number
            * *sort* (``aiosu.models.common.SortTypes``) --
                Optional, sort order, defaults to ``id_desc``
            * *types* (``list[str]``) --
                Optional, post types
            * *user_id* (``int``) --
                Optional, user ID
            * with_deleted (``bool``) --
                Optional, include deleted discussions
            * cursor_string (``str``) --
                Optional, cursor string

        :raises APIException: Contains status code and error message
        :return: Beatmapset discussion post response
        :rtype: aiosu.models.beatmap.BeatmapsetDiscussionPostResponse
        """
        url = f"{self.base_url}/api/v2/beatmapsets/discussions/posts"
        params: dict[str, Any] = {}
        add_param(params, kwargs, key="beatmapset_discussion_id")
        add_param(params, kwargs, key="limit")
        add_param(params, kwargs, key="page")
        add_param(params, kwargs, key="sort")
        add_param(params, kwargs, key="types")
        add_param(params, kwargs, key="user", param_name="user_id")
        add_param(params, kwargs, key="with_deleted", converter=to_lower_str)
        add_param(params, kwargs, key="cursor_string")
        json = await self._request("GET", url, params=params)
        resp = BeatmapsetDiscussionPostResponse.model_validate(json)
        if resp.cursor_string:
            kwargs["cursor_string"] = resp.cursor_string
            resp.next = partial(self.get_beatmapset_discussion_posts, **kwargs)
        return resp

    @prepare_token
    @check_token
    @requires_scope(Scopes.PUBLIC)
    async def get_beatmapset_discussion_votes(
        self,
        **kwargs: Any,
    ) -> BeatmapsetDiscussionVoteResponse:
        r"""Get beatmapset discussion votes.

        :param \**kwargs:
            See below

        :Keyword Arguments:
            * *beatmapset_discussion_id* (``int``) --
                Optional, beatmapset discussion ID
            * *limit* (``int``) --
                Optional, number of results per page
            * *page* (``int``) --
                Optional, page number
            * *receiver_id* (``int``) --
                Optional, receiver ID
            * *score* (``aiosu.models.beatmap.BeatmapsetDiscussionVoteScore``) --
                Optional, vote score
            * *sort* (``aiosu.models.common.SortTypes``) --
                Optional, sort order, defaults to ``id_desc``
            * *user_id* (``int``) --
                Optional, user ID
            * with_deleted (``bool``) --
                Optional, include deleted discussions
            * cursor_string (``str``) --
                Optional, cursor string

        :raises APIException: Contains status code and error message
        :return: Beatmapset discussion vote response
        :rtype: aiosu.models.beatmap.BeatmapsetDiscussionVoteResponse
        """
        url = f"{self.base_url}/api/v2/beatmapsets/discussions/votes"
        params: dict[str, Any] = {}
        add_param(params, kwargs, key="beatmapset_discussion_id")
        add_param(params, kwargs, key="limit")
        add_param(params, kwargs, key="page")
        add_param(params, kwargs, key="receiver", param_name="receiver_id")
        add_param(params, kwargs, key="score")
        add_param(params, kwargs, key="sort")
        add_param(params, kwargs, key="user", param_name="user_id")
        add_param(params, kwargs, key="with_deleted", converter=to_lower_str)
        add_param(params, kwargs, key="cursor_string")
        json = await self._request("GET", url, params=params)
        resp = BeatmapsetDiscussionVoteResponse.model_validate(json)
        if resp.cursor_string:
            kwargs["cursor_string"] = resp.cursor_string
            resp.next = partial(self.get_beatmapset_discussion_votes, **kwargs)
        return resp

    @prepare_token
    @check_token
    @requires_scope(Scopes.PUBLIC)
    async def get_score(
        self,
        score_id: int,
        mode: Gamemode,
    ) -> Score:
        r"""Gets data about a score.

        :param score_id: The ID of the score
        :type score_id: int
        :param mode: The gamemode to search for
        :type mode: aiosu.models.gamemode.Gamemode

        :raises APIException: Contains status code and error message
        :return: Score data object
        :rtype: aiosu.models.score.Score
        """
        url = f"{self.base_url}/api/v2/scores/{mode}/{score_id}"
        json = await self._request("GET", url)
        return Score.model_validate(json)

    @prepare_token
    @check_token
    @requires_scope(Scopes.PUBLIC)
    @requires_scope(Scopes.IDENTIFY | Scopes.DELEGATE, any_scope=True)
    async def get_score_replay(
        self,
        score_id: int,
        mode: Gamemode,
    ) -> BytesIO:
        r"""Gets the replay file for a score.

        :param score_id: The ID of the score
        :type score_id: int
        :param mode: The gamemode to search for
        :type mode: aiosu.models.gamemode.Gamemode

        :raises APIException: Contains status code and error message
        :return: Replay file
        :rtype: io.BytesIO
        """
        url = f"{self.base_url}/api/v2/scores/{mode}/{score_id}/download"
        return await self._request("GET", url)

    @prepare_token
    @check_token
    @requires_scope(Scopes.PUBLIC)
    async def get_rankings(
        self,
        mode: Gamemode,
        type: RankingType,
        **kwargs: Any,
    ) -> Rankings:
        r"""Get rankings.

        :param mode: The gamemode to search for
        :type mode: aiosu.models.gamemode.Gamemode
        :param type: The ranking type to search for
        :type type: aiosu.models.rankings.RankingType
        :param \**kwargs:
            See below

        :Keyword Arguments:
            * *country* (``str``) --
                Optional, country code
            * *filter* (``aiosu.models.rankings.RankingFilter``) --
                Optional, ranking filter
            * *spotlight* (``int``) --
                Optional, spotlight ID
            * *variant* (``aiosu.models.rankings.RankingVariant``) --
                Optional, ranking variant
            * *cursor_string* (``str``) --
                Optional, cursor string

        :raises APIException: Contains status code and error message
        :return: Rankings
        :rtype: aiosu.models.rankings.Rankings
        """
        url = f"{self.base_url}/api/v2/rankings/{mode}/{type}"
        params: dict[str, Any] = {}
        add_param(params, kwargs, key="country")
        add_param(params, kwargs, key="filter")
        add_param(params, kwargs, key="spotlight")
        add_param(params, kwargs, key="variant")
        add_param(params, kwargs, key="cursor_string")
        json = await self._request("GET", url, params=params)
        resp = Rankings.model_validate(json)
        if resp.cursor_string:  # Unused: API does not return cursor_string
            kwargs["cursor_string"] = resp.cursor_string
            resp.next = partial(self.get_rankings, mode=mode, type=type, **kwargs)
        return resp

    @prepare_token
    @check_token
    @requires_scope(Scopes.PUBLIC)
    async def get_spotlights(self) -> list[Spotlight]:
        r"""Gets the current spotlights.

        :raises APIException: Contains status code and error message
        :return: List of spotlights
        :rtype: list[aiosu.models.spotlight.Spotlight]
        """
        url = f"{self.base_url}/api/v2/spotlights"
        json = await self._request("GET", url)
        return from_list(Spotlight.model_validate, json.get("spotlights", []))

    @prepare_token
    @check_token
    @requires_scope(Scopes.PUBLIC)
    async def get_forum_topic(self, topic_id: int, **kwargs: Any) -> ForumTopicResponse:
        r"""Gets a forum topic.

        :param topic_id: The ID of the topic
        :type topic_id: int
        :param \**kwargs:
            See below

        :Keyword Arguments:
            * *limit* (``int``) --
                Optional, the number of posts to return. Min: 1, Max: 50, defaults to 20
            * *sort* (``aiosu.models.common.SortTypes``) --
                Optional, the sort type to use. Defaults to ``id_asc``
            * *start* (``int``) --
                Optional, the start post ID to use for pagination.
            * *end* (``int``) --
                Optional, the end post ID to use for pagination.
            * *cursor_string* (``str``) --
                Optional, the cursor string to use for pagination.

        :raises APIException: Contains status code and error message
        :return: Forum topic response object
        :rtype: aiosu.models.forum.ForumTopicResponse
        """
        if not 1 <= (limit := kwargs.pop("limit", 20)) <= 50:
            raise ValueError("Invalid limit specified. Limit must be between 1 and 50")
        url = f"{self.base_url}/api/v2/forums/topics/{topic_id}"
        params: dict[str, Any] = {
            "limit": limit,
        }
        add_param(params, kwargs, key="sort")
        add_param(params, kwargs, key="start")
        add_param(params, kwargs, key="end")
        add_param(params, kwargs, key="cursor_string")
        json = await self._request("GET", url, params=params)
        resp = ForumTopicResponse.model_validate(json)
        if resp.cursor_string:
            kwargs["cursor_string"] = resp.cursor_string
            resp.next = partial(self.get_forum_topic, topic_id, **kwargs)
        return resp

    @prepare_token
    @check_token
    @requires_scope(Scopes.FORUM_WRITE)
    @requires_scope(Scopes.IDENTIFY | Scopes.DELEGATE, any_scope=True)
    async def create_forum_topic(
        self,
        forum_id: int,
        title: str,
        content: str,
        **kwargs: Any,
    ) -> ForumCreateTopicResponse:
        r"""Creates a forum topic.

        :param forum_id: The ID of the forum to create the topic in
        :type forum_id: int
        :param title: The title of the topic
        :type title: str
        :param content: The content of the topic
        :type content: str
        :param \**kwargs:
            See below

        :Keyword Arguments:
            * *with_poll* (``bool``) --
                Optional, whether to create a poll with the topic. Defaults to ``False``
            * *poll_title* (``str``) --
                Optional, the title of the poll
            * *poll_options* (``list[str]``) --
                Optional, the options for the poll
            * *poll_length_days* (``int``) --
                Optional, the length of the poll in days. Defaults to 0
            * *poll_vote_change* (``bool``) --
                Optional, whether to allow users to change their vote. Defaults to ``False``
            * *poll_hide_results* (``bool``) --
                Optional, whether to hide the results of the poll. Defaults to ``False``
            * *poll_max_votes* (``int``) --
                Optional, the maximum number of votes a user can make. Defaults to 1

        :raises APIException: Contains status code and error message
        :return: Forum create topic response object
        :rtype: aiosu.models.forum.ForumCreateTopicResponse
        """
        url = f"{self.base_url}/api/v2/forums/topics"
        data: dict[str, Any] = {
            "forum_id": forum_id,
            "title": title,
            "body": content,
        }
        add_param(data, kwargs, key="with_poll")
        if data.get("with_poll"):
            forum_topic_poll: dict[str, Any] = {
                "title": kwargs["poll_title"],
                "length_days": kwargs.pop("poll_length_days", 0),
                "vote_change": kwargs.pop("poll_vote_change", False),
                "hide_results": kwargs.pop("poll_hide_results", False),
                "max_votes": kwargs.pop("poll_max_votes", 1),
            }
            add_param(
                forum_topic_poll,
                kwargs,
                key="options",
                param_name="poll_options",
            )
            data["forum_topic_poll"] = forum_topic_poll
        json = await self._request("POST", url, json=data)
        return ForumCreateTopicResponse.model_validate(json)

    @prepare_token
    @check_token
    @requires_scope(Scopes.FORUM_WRITE)
    @requires_scope(Scopes.IDENTIFY | Scopes.DELEGATE, any_scope=True)
    async def reply_forum_topic(self, topic_id: int, content: str) -> ForumPost:
        r"""Replies to a forum topic.

        :param topic_id: The ID of the topic
        :type topic_id: int
        :param content: The content of the post
        :type content: str
        :raises APIException: Contains status code and error message
        :return: Forum post object
        :rtype: aiosu.models.forum.ForumPost
        """
        url = f"{self.base_url}/api/v2/forums/topics/{topic_id}/reply"
        data: dict[str, str] = {
            "body": content,
        }
        json = await self._request("POST", url, json=data)
        return ForumPost.model_validate(json)

    @prepare_token
    @check_token
    @requires_scope(Scopes.FORUM_WRITE)
    async def edit_forum_topic_title(self, topid_id: int, new_title: str) -> ForumTopic:
        r"""Edits a forum topic's title.

        :param topid_id: The ID of the topic
        :type topid_id: int
        :param new_title: The new title of the topic
        :type new_title: str
        :raises APIException: Contains status code and error message
        :return: Forum topic object
        :rtype: aiosu.models.forum.ForumTopic
        """
        url = f"{self.base_url}/api/v2/forums/topics/{topid_id}/title"
        data: dict[str, dict[str, str]] = {
            "forum_topic": {
                "topic_title": new_title,
            },
        }
        json = await self._request("PUT", url, json=data)
        return ForumTopic.model_validate(json)

    @prepare_token
    @check_token
    @requires_scope(Scopes.FORUM_WRITE)
    async def edit_forum_post(self, post_id: int, new_content: str) -> ForumPost:
        r"""Edits a forum post.

        :param post_id: The ID of the post
        :type post_id: int
        :param new_content: The new content of the post
        :type new_content: str
        :raises APIException: Contains status code and error message
        :return: Forum post object
        :rtype: aiosu.models.forum.ForumPost
        """
        url = f"{self.base_url}/api/v2/forums/posts/{post_id}"
        data: dict[str, str] = {
            "body": new_content,
        }
        json = await self._request("PUT", url, json=data)
        return ForumPost.model_validate(json)

    @prepare_token
    @check_token
    @requires_scope(Scopes.LAZER)
    @requires_scope(Scopes.IDENTIFY | Scopes.DELEGATE, any_scope=True)
    async def get_chat_ack(self, **kwargs: Any) -> list[ChatUserSilence]:
        r"""Gets chat acknowledgement.

        :param \**kwargs:
            See below

        :Keyword Arguments:
            * *since* (``int``) --
                Optional, the last message ID received
            * *silence_id_since* (``int``) --
                Optional, the last silence ID received

        :raises APIException: Contains status code and error message
        :return: List of chat user silence objects
        :rtype: list[aiosu.models.chat.ChatUserSilence]
        """
        url = f"{self.base_url}/api/v2/chat/ack"
        data: dict[str, Any] = {}
        add_param(data, kwargs, key="since")
        add_param(data, kwargs, key="silence_id_since", param_name="history_since")
        json = await self._request("POST", url, json=data)
        return from_list(ChatUserSilence.model_validate, json.get("silences", []))

    @prepare_token
    @check_token
    @requires_scope(Scopes.LAZER)
    @requires_scope(Scopes.IDENTIFY | Scopes.DELEGATE, any_scope=True)
    async def get_chat_updates(self, since: int, **kwargs: Any) -> ChatUpdateResponse:
        r"""Gets chat updates.

        :param since: The last message ID received
        :type since: int
        :param \**kwargs:
            See below

        :Keyword Arguments:
            * *limit* (``int``) --
                Optional, the maximum number of messages to return. Min: 1, Max: 50. Defaults to 50
            * *channel_id* (``int``) --
                Optional, the channel ID to get messages from
            * *silence_id_since* (``int``) --
                Optional, the last silence ID received
            * *includes* (``list[aiosu.models.chat.ChatIncludeTypes]``) --
                Optional, the additional information to include. Defaults to all.

        :raises ValueError: If limit is not between 1 and 50
        :raises APIException: Contains status code and error message
        :return: Chat update response object
        :rtype: aiosu.models.chat.ChatUpdateResponse
        """
        if not 1 <= (limit := kwargs.get("limit", 50)) <= 50:
            raise ValueError("limit must be between 1 and 50")
        url = f"{self.base_url}/api/v2/chat/updates"
        params: dict[str, Any] = {
            "since": since,
            "limit:": limit,
        }
        add_param(params, kwargs, key="channel_id")
        add_param(params, kwargs, key="includes")
        add_param(params, kwargs, key="silence_id_since", param_name="history_since")
        json = await self._request("GET", url, params=params)
        return ChatUpdateResponse.model_validate(json)

    @prepare_token
    @check_token
    @requires_scope(Scopes.LAZER)
    @requires_scope(Scopes.IDENTIFY | Scopes.DELEGATE, any_scope=True)
    async def get_channel(self, channel_id: int) -> ChatChannelResponse:
        r"""Gets channel.

        :param channel_id: The channel ID to get
        :type channel_id: int
        :raises APIException: Contains status code and error message
        :return: Chat channel object
        :rtype: aiosu.models.chat.ChatChannelResponse
        """
        url = f"{self.base_url}/api/v2/chat/channels/{channel_id}"
        json = await self._request("GET", url)
        return ChatChannelResponse.model_validate(json)

    @prepare_token
    @check_token
    @requires_scope(Scopes.LAZER)
    @requires_scope(Scopes.IDENTIFY | Scopes.DELEGATE, any_scope=True)
    async def get_channels(self) -> list[ChatChannel]:
        r"""Gets a list of joinable public channels.

        :raises APIException: Contains status code and error message
        :return: List of chat channel objects
        :rtype: list[aiosu.models.chat.ChatChannel]
        """
        url = f"{self.base_url}/api/v2/chat/channels"
        json = await self._request("GET", url)
        return from_list(ChatChannel.model_validate, json)

    @prepare_token
    @check_token
    @requires_scope(Scopes.LAZER)
    @requires_scope(Scopes.IDENTIFY | Scopes.DELEGATE, any_scope=True)
    async def get_channel_messages(
        self,
        channel_id: int,
        **kwargs: Any,
    ) -> list[ChatMessage]:
        r"""Gets channel messages.

        :param channel_id: The channel ID to get messages from
        :type channel_id: int
        :param \**kwargs:
            See below

        :Keyword Arguments:
            * *limit* (``int``) --
                Optional, the maximum number of messages to return. Min: 1, Max: 50. Defaults to 50
            * *since* (``int``) --
                Optional, the ID of the oldest message to return
            * *until* (``int``) --
                Optional, the ID of the newest message to return

        :raises ValueError: If limit is not between 1 and 50
        :raises APIException: Contains status code and error message
        :return: List of chat message objects
        :rtype: list[aiosu.models.chat.ChatMessage]
        """
        if not 1 <= (limit := kwargs.get("limit", 50)) <= 50:
            raise ValueError("limit must be between 1 and 50")
        url = f"{self.base_url}/api/v2/chat/channels/{channel_id}/messages"
        params: dict[str, Any] = {
            "limit:": limit,
        }
        add_param(params, kwargs, key="since")
        add_param(params, kwargs, key="until")
        json = await self._request("GET", url, params=params)
        return from_list(ChatMessage.model_validate, json)

    @prepare_token
    @check_token
    @requires_scope(Scopes.CHAT_WRITE)
    @requires_scope(Scopes.IDENTIFY | Scopes.DELEGATE, any_scope=True)
    async def create_chat_channel(
        self,
        type: ChatChannelTypes,
        **kwargs: Any,
    ) -> ChatChannel:
        r"""Creates a chat channel.

        :param type: The type of the channel.
        :type type: aiosu.models.chat.ChatChannelType
        :param \**kwargs:
            See below

        :Keyword Arguments:
            * *message* (``str``) --
                Required if type is ``ANNOUNCE``, the message to send in the PM
            * *target_id* (``int``) --
                Only used if if type is ``PM``, the ID of the user to send a PM to
            * *target_ids* (``List[int]``) --
                Only used if type is ``ANNOUNCE``, the IDs of the users to send a PM to
            * *channel_name* (``str``) --
                Only used if type is ``ANNOUNCE``, the name of the channel
            * *channel_description* (``str``) --
                Only used if type is ``ANNOUNCE``, the description of the channel

        :raises ValueError: If missing required parameters
        :raises APIException: Contains status code and error message
        :return: Chat channel object
        :rtype: aiosu.models.chat.ChatChannel
        """
        url = f"{self.base_url}/api/v2/chat/channels"
        data: dict[str, Any] = {
            "type": type,
        }
        add_param(data, kwargs, key="message")
        if type == "PM":
            if not add_param(data, kwargs, key="target_id"):
                raise ValueError("Missing target ID")
        elif type == "ANNOUNCE":
            if not add_param(data, kwargs, key="target_ids"):
                raise ValueError("Missing target IDs")
            if not data.get("message"):
                raise ValueError("Missing message")
            channel = {
                "name": kwargs["channel_name"],
                "description": kwargs["channel_description"],
            }
            data["channel"] = channel
        json = await self._request("POST", url, json=data)
        return ChatChannel.model_validate(json)

    @prepare_token
    @check_token
    @requires_scope(Scopes.LAZER)
    @requires_scope(Scopes.IDENTIFY | Scopes.DELEGATE, any_scope=True)
    async def join_channel(self, channel_id: int, user_id: int) -> ChatChannel:
        r"""Joins a channel.

        :param channel_id: The channel ID to join
        :type channel_id: int
        :param user_id: The user ID to join as
        :type user_id: int
        :raises APIException: Contains status code and error message
        :return: Chat channel object
        :rtype: aiosu.models.chat.ChatChannel
        """
        url = f"{self.base_url}/api/v2/chat/channels/{channel_id}/users/{user_id}"
        json = await self._request("PUT", url)
        return ChatChannel.model_validate(json)

    @prepare_token
    @check_token
    @requires_scope(Scopes.LAZER)
    @requires_scope(Scopes.IDENTIFY | Scopes.DELEGATE, any_scope=True)
    async def leave_channel(self, channel_id: int, user_id: int) -> None:
        r"""Leaves a channel.

        :param channel_id: The channel ID to leave
        :type channel_id: int
        :param user_id: The user ID to leave as
        :type user_id: int
        :raises APIException: Contains status code and error message
        """
        url = f"{self.base_url}/api/v2/chat/channels/{channel_id}/users/{user_id}"
        await self._request("DELETE", url)

    @prepare_token
    @check_token
    @requires_scope(Scopes.LAZER)
    @requires_scope(Scopes.IDENTIFY | Scopes.DELEGATE, any_scope=True)
    async def mark_read(self, channel_id: int, message_id: int) -> None:
        r"""Marks a channel as read.

        :param channel_id: The channel ID to mark as read
        :type channel_id: int
        :param message_id: The message ID to mark as read up to
        :type message_id: int
        :raises APIException: Contains status code and error message
        """
        url = f"{self.base_url}/api/v2/chat/channels/{channel_id}/mark-as-read/{message_id}"
        await self._request("PUT", url)

    @prepare_token
    @check_token
    @requires_scope(Scopes.LAZER)
    @requires_scope(Scopes.IDENTIFY | Scopes.DELEGATE, any_scope=True)
    async def send_message(
        self,
        channel_id: int,
        message: str,
        is_action: bool,
    ) -> ChatMessage:
        r"""Sends a message to a channel.

        :param channel_id: The ID of the channel
        :type channel_id: int
        :param message: The message to send
        :type message: str
        :param is_action: Whether the message is an action
        :type is_action: bool
        :raises APIException: Contains status code and error message
        :return: Chat message object
        :rtype: aiosu.models.chat.ChatMessage
        """
        url = f"{self.base_url}/api/v2/chat/channels/{channel_id}/messages"
        data: dict[str, Any] = {
            "message": message,
            "is_action": is_action,
        }
        json = await self._request("POST", url, json=data)
        return ChatMessage.model_validate(json)

    @prepare_token
    @check_token
    @requires_scope(Scopes.CHAT_WRITE)
    @requires_scope(Scopes.IDENTIFY | Scopes.DELEGATE, any_scope=True)
    async def send_private_message(
        self,
        user_id: int,
        message: str,
        is_action: bool,
        **kwargs: Any,
    ) -> ChatMessageCreateResponse:
        r"""Sends a message to a user.

        :param user_id: The ID of the user
        :type user_id: int
        :param message: The message to send
        :type message: str
        :param is_action: Whether the message is an action
        :type is_action: bool
        :param \**kwargs:
            See below

        :Keyword Arguments:
            * *uuid* (``str``) --
                Optional, client-side message identifier to be sent back in response and websocket json

        :raises APIException: Contains status code and error message
        :return: Chat message create response object
        :rtype: aiosu.models.chat.ChatMessageCreateResponse
        """
        url = f"{self.base_url}/api/v2/chat/new"
        data: dict[str, Any] = {
            "target_id": user_id,
            "message": message,
            "is_action": is_action,
        }
        add_param(data, kwargs, key="uuid")
        json = await self._request("POST", url, json=data)
        return ChatMessageCreateResponse.model_validate(json)

    @prepare_token
    @check_token
    @requires_scope(Scopes.PUBLIC)
    async def get_multiplayer_matches(
        self,
        **kwargs: Any,
    ) -> MultiplayerMatchesResponse:
        r"""Gets the multiplayer matches.

        :param \**kwargs:
            See below

        :Keyword Arguments:
            * *sort* (``aiosu.models.common.SortTypes``) --
                Optional, the sort type
            * *limit* (``int``) --
                Optional, number of scores to get. Min: 1, Max: 50, defaults to 50
            * *cursor_string* (``str``) --
                Optional, the cursor string to get the next page of results

        :raises ValueError: If limit is not between 1 and 50
        :raises APIException: Contains status code and error message
        :return: Multiplayer matches response object
        :rtype: aiosu.models.multiplayer.MultiplayerMatchesResponse
        """
        if not 1 <= (limit := kwargs.pop("limit", 1)) <= 50:
            raise ValueError("Limit must be between 1 and 50")
        url = f"{self.base_url}/api/v2/matches"
        params: dict[str, Any] = {
            "limit": limit,
        }
        add_param(params, kwargs, key="sort")
        json = await self._request("GET", url, params=params)
        resp = MultiplayerMatchesResponse.model_validate(json)
        if resp.cursor_string:
            kwargs["cursor_string"] = resp.cursor_string
            resp.next = partial(self.get_multiplayer_matches, **kwargs)
        return resp

    @prepare_token
    @check_token
    @requires_scope(Scopes.PUBLIC)
    async def get_multiplayer_match(
        self,
        match_id: int,
        **kwargs: Any,
    ) -> MultiplayerMatchResponse:
        r"""Gets a multiplayer match.

        :param match_id: The ID of the match
        :type match_id: int
        :param \**kwargs:
            See below

        :Keyword Arguments:
            * *limit* (``int``) --
                Optional, number of scores to get. Min: 1, Max: 100, defaults to 100
            * *before* (``int``) --
                Optional, the ID of the score to get the scores before
            * *after* (``int``) --
                Optional, the ID of the score to get the scores after

        :raises ValueError: If limit is not between 1 and 100
        :raises APIException: Contains status code and error message
        :return: Multiplayer match response object
        :rtype: aiosu.models.multiplayer.MultiplayerMatchResponse
        """
        if not 1 <= (limit := kwargs.pop("limit", 1)) <= 100:
            raise ValueError("Limit must be between 1 and 100")
        url = f"{self.base_url}/api/v2/matches/{match_id}"
        params: dict[str, Any] = {
            "limit": limit,
        }
        add_param(params, kwargs, key="before")
        add_param(params, kwargs, key="after")
        json = await self._request("GET", url)
        return MultiplayerMatchResponse.model_validate(json)

    @prepare_token
    @check_token
    @requires_scope(Scopes.PUBLIC)
    @requires_scope(Scopes.IDENTIFY | Scopes.DELEGATE, any_scope=True)
    async def get_multiplayer_rooms(self, **kwargs: Any) -> list[MultiplayerRoom]:
        r"""Gets the multiplayer rooms.

        :param \**kwargs:
            See below

        :Keyword Arguments:
            * *mode* (``aiosu.models.multiplayer.MultiplayerRoomMode``) --
                Optional, the multiplayer room mode
            * *limit* (``int``) --
                Optional, number of scores to get. Min: 1, Max: 50, defaults to 50
            * *sort* (``aiosu.models.common.SortTypes``) --
                Optional, the sort type
            * *category* (``aiosu.models.multiplayer.MultiplayerRoomCategories``) --
                Optional, the multiplayer room category
            * *type* (``aiosu.models.multiplayer.MultiplayerRoomTypeGroups``) --
                Optional, the multiplayer room type group

        :raises ValueError: If limit is not between 1 and 50
        :raises APIException: Contains status code and error message
        :return: List of multiplayer rooms
        :rtype: list[aiosu.models.multiplayer.MultiplayerRoom]
        """
        if not 1 <= (limit := kwargs.pop("limit", 50)) <= 50:
            raise ValueError("Limit must be between 1 and 50")
        url = f"{self.base_url}/api/v2/rooms"
        if "mode" in kwargs:
            mode: MultiplayerRoomMode = kwargs.pop("mode")
            url += f"/{mode}"
        params: dict[str, Any] = {
            "limit": limit,
        }
        add_param(params, kwargs, key="sort")
        add_param(params, kwargs, key="category")
        add_param(params, kwargs, key="type", param_name="type_group")
        json = await self._request("GET", url, params=params)
        return from_list(MultiplayerRoom.model_validate, json)

    @prepare_token
    @check_token
    @requires_scope(Scopes.PUBLIC)
    async def get_multiplayer_room(self, room_id: int) -> MultiplayerRoom:
        r"""Gets a multiplayer room.

        :param room_id: The ID of the room
        :type room_id: int

        :raises APIException: Contains status code and error message
        :return: Multiplayer room object
        :rtype: aiosu.models.multiplayer.MultiplayerRoom
        """
        url = f"{self.base_url}/api/v2/rooms/{room_id}"
        json = await self._request("GET", url)
        return MultiplayerRoom.model_validate(json)

    @prepare_token
    @check_token
    @requires_scope(Scopes.PUBLIC)
    @requires_scope(Scopes.IDENTIFY | Scopes.DELEGATE, any_scope=True)
    async def get_multiplayer_leaderboard(
        self,
        room_id: int,
        **kwargs: Any,
    ) -> MultiplayerLeaderboardResponse:
        r"""Gets the multiplayer leaderboard for a room.

        :param room_id: The ID of the room
        :type room_id: int
        :param \**kwargs:
            See below

        :Keyword Arguments:
            * *limit* (``int``) --
                Optional, number of scores to get. Min: 1, Max: 50, defaults to 50

        :raises ValueError: If limit is not between 1 and 50
        :raises APIException: Contains status code and error message
        :return: Multiplayer leaderboard response object
        :rtype: aiosu.models.multiplayer.MultiplayerLeaderboardResponse
        """
        if not 1 <= (limit := kwargs.pop("limit", 50)) <= 50:
            raise ValueError("Limit must be between 1 and 50")
        url = f"{self.base_url}/api/v2/rooms/{room_id}/leaderboard"
        params: dict[str, Any] = {
            "limit": limit,
        }
        json = await self._request("GET", url, params=params)
        return MultiplayerLeaderboardResponse.model_validate(json)

    @prepare_token
    @check_token
    @requires_scope(Scopes.PUBLIC)
    async def get_multiplayer_scores(
        self,
        room_id: int,
        playlist_id: int,
        **kwargs: Any,
    ) -> MultiplayerScoresResponse:
        r"""Gets the multiplayer scores for a room.

        :param room_id: The ID of the room
        :type room_id: int
        :param playlist_id: The ID of the playlist
        :type playlist_id: int
        :param \**kwargs:
            See below

        :Keyword Arguments:
            * *limit* (``int``) --
                Optional, the number of scores to return
            * *sort* (``aiosu.models.multiplayer.MultiplayerScoreSortType``) --
                Optional, the sort order of the scores
            * *cursor_string* (``str``) --
                Optional, the cursor string to use for pagination

        :raises ValueError: If limit is not between 1 and 100
        :raises APIException: Contains status code and error message
        :return: Multiplayer scores response object
        :rtype: aiosu.models.multiplayer.MultiplayerScoresResponse
        """
        if not 1 <= (limit := kwargs.pop("limit", 1)) <= 100:
            raise ValueError("Limit must be between 1 and 100")
        url = f"{self.base_url}/api/v2/rooms/{room_id}/playlist/{playlist_id}/scores"
        params: dict[str, Any] = {
            "limit": limit,
        }
        add_param(params, kwargs, key="sort")
        add_param(params, kwargs, key="cursor_string")
        json = await self._request("GET", url, params=params)
        resp = MultiplayerScoresResponse.model_validate(json)
        if resp.cursor_string:
            kwargs["cursor_string"] = resp.cursor_string
            resp.next = partial(
                self.get_multiplayer_scores,
                room_id,
                playlist_id,
                **kwargs,
            )
        return resp

    @prepare_token
    @check_token
    async def revoke_token(self) -> None:
        r"""Revokes the current token and closes the session.

        :raises APIException: Contains status code and error message
        """
        url = f"{self.base_url}/api/v2/oauth/tokens/current"
        await self._request("DELETE", url)
        await self._delete_token()
        await self.close()

    async def close(self) -> None:
        """Closes the client session."""
        if self._session:
            await self._session.close()
            self._session = None



"""
This module contains models for API v2 scopes.
"""
from __future__ import annotations

from enum import IntFlag
from enum import unique

__all__ = (
    "Scopes",
    "VALID_CLIENT_SCOPES",
)


@unique
class Scopes(IntFlag):
    NONE = 0
    PUBLIC = 1 << 0
    IDENTIFY = 1 << 1
    FRIENDS_READ = 1 << 2
    FORUM_WRITE = 1 << 3
    DELEGATE = 1 << 4
    CHAT_WRITE = 1 << 5
    LAZER = 1 << 6  # unused, lazer endpoints are not planned for support

    def __flags__(self) -> list[Scopes]:
        scopes_list = [scope for scope in Scopes if self & scope]
        return scopes_list

    def __str__(self) -> str:
        return " ".join(
            scope_name
            for scope_name, scope in API_SCOPE_NAMES.items()
            if scope in self.__flags__()
        )

    @classmethod
    def from_api_list(cls, scopes: list[str]) -> Scopes:
        return cls(sum(API_SCOPE_NAMES[scope] for scope in scopes))


API_SCOPE_NAMES = {
    "public": Scopes.PUBLIC,
    "identify": Scopes.IDENTIFY,
    "friends.read": Scopes.FRIENDS_READ,
    "forum.write": Scopes.FORUM_WRITE,
    "delegate": Scopes.DELEGATE,
    "chat.write": Scopes.CHAT_WRITE,
    "lazer": Scopes.LAZER,
}
VALID_CLIENT_SCOPES = Scopes.PUBLIC | Scopes.DELEGATE



"""
This module contains accuracy calculators for osu! gamemodes.
"""
from __future__ import annotations

import abc
from typing import TYPE_CHECKING

from ..models.gamemode import Gamemode
from ..models.mods import Mod

if TYPE_CHECKING:
    from typing import Type
    from typing import Callable
    from ..models.score import Score

__all__ = [
    "OsuAccuracyCalculator",
    "TaikoAccuracyCalculator",
    "ManiaAccuracyCalculator",
    "CatchAccuracyCalculator",
]

cast_int: Callable[..., int] = lambda x: int(x or 0)


def get_calculator(mode: Gamemode) -> Type[AbstractAccuracyCalculator]:
    r"""Returns the accuracy calculator for the given gamemode.

    :param mode: The gamemode to get the calculator for
    :type mode: aiosu.models.gamemode.Gamemode
    :raises ValueError: If the gamemode is unknown
    :return: The accuracy calculator type for the given gamemode
    :rtype: Type[AbstractAccuracyCalculator]
    """
    if mode == Gamemode.STANDARD:
        return OsuAccuracyCalculator
    elif mode == Gamemode.TAIKO:
        return TaikoAccuracyCalculator
    elif mode == Gamemode.MANIA:
        return ManiaAccuracyCalculator
    elif mode == Gamemode.CTB:
        return CatchAccuracyCalculator
    else:
        raise ValueError(f"Unknown gamemode: {mode}")


class AbstractAccuracyCalculator(abc.ABC):
    @staticmethod
    @abc.abstractmethod
    def calculate(score: Score) -> float:
        ...

    @staticmethod
    @abc.abstractmethod
    def calculate_weighted(score: Score) -> float:
        ...


class OsuAccuracyCalculator(AbstractAccuracyCalculator):
    @staticmethod
    def calculate(score: Score) -> float:
        r"""Calculates accuracy for an osu!std score.

        :param score: The score to calculate accuracy for
        :type score: aiosu.models.score.Score
        :return: Accuracy for the given score
        :rtype: float
        """
        total_hits = (
            score.statistics.count_300
            + score.statistics.count_100
            + score.statistics.count_50
            + score.statistics.count_miss
        )

        if total_hits <= 0:
            return 0.0

        return max(
            (
                score.statistics.count_300 * 6.0
                + score.statistics.count_100 * 2.0
                + score.statistics.count_50
            )
            / (total_hits * 6.0),
            0.0,
        )

    @staticmethod
    def calculate_weighted(score: Score) -> float:
        r"""Calculates weighted accuracy for an osu!std score.

        :param score: The score to calculate accuracy for
        :type score: aiosu.models.score.Score
        :raises ValueError: If score does not have an associated beatmap
        :return: Weighted accuracy to be used in pp calculation
        :rtype: float
        """
        if score.beatmap is None:
            raise ValueError("Given score does not have a beatmap.")

        total_hits = (
            score.statistics.count_300
            + score.statistics.count_100
            + score.statistics.count_50
            + score.statistics.count_miss
        )

        amount_hit_objects_with_accuracy = cast_int(score.beatmap.count_circles)

        if Mod.ScoreV2 in score.mods:  # TODO: Check for lazer classic mod
            amount_hit_objects_with_accuracy += cast_int(score.beatmap.count_sliders)

        if amount_hit_objects_with_accuracy <= 0:
            return 0.0

        return max(
            (
                (
                    score.statistics.count_300
                    - (total_hits - amount_hit_objects_with_accuracy)
                )
                * 6.0
                + score.statistics.count_100 * 2.0
                + score.statistics.count_50
            )
            / (amount_hit_objects_with_accuracy * 6.0),
            0.0,
        )


class TaikoAccuracyCalculator(AbstractAccuracyCalculator):
    @staticmethod
    def calculate(score: Score) -> float:
        r"""Calculates accuracy for an osu!taiko score.

        :param score: The score to calculate accuracy for
        :type score: aiosu.models.score.Score
        :return: Accuracy for the given score
        :rtype: float
        """
        total_hits = (
            score.statistics.count_300
            + score.statistics.count_100
            + score.statistics.count_50
            + score.statistics.count_miss
        )

        if total_hits <= 0:
            return 0.0

        return max(
            (score.statistics.count_300 * 2.0 + score.statistics.count_100)
            / (total_hits * 2.0),
            0.0,
        )

    @classmethod
    def calculate_weighted(cls, score: Score) -> float:
        r"""Calculates weighted accuracy for an osu!taiko score.

        :param score: The score to calculate accuracy for
        :type score: aiosu.models.score.Score
        :return: Weighted accuracy to be used in pp calculation
        :rtype: float
        """
        return cls.calculate(score)


class ManiaAccuracyCalculator(AbstractAccuracyCalculator):
    @staticmethod
    def calculate(score: Score) -> float:
        r"""Calculates accuracy for an osu!mania score.

        :param score: The score to calculate accuracy for
        :type score: aiosu.models.score.Score
        :return: Accuracy for the given score
        :rtype: float
        """
        count_perfect = score.statistics.count_geki
        count_great = score.statistics.count_300
        count_good = score.statistics.count_katu
        count_ok = score.statistics.count_100
        count_meh = score.statistics.count_50
        count_miss = score.statistics.count_miss

        total_hits = (
            count_perfect + count_ok + count_great + count_good + count_meh + count_miss
        )

        if total_hits <= 0:
            return 0.0

        if Mod.ScoreV2 in score.mods:
            return max(
                (
                    +(count_perfect * 305)
                    + (count_great * 300)
                    + (count_good * 200)
                    + (count_ok * 100)
                    + (count_meh * 50)
                )
                / (total_hits * 305),
                0.0,
            )

        return max(
            (
                +((count_perfect + count_great) * 6.0)
                + (count_good * 4.0)
                + (count_ok * 2.0)
                + count_meh
            )
            / (total_hits * 6.0),
            0.0,
        )

    @staticmethod
    def calculate_weighted(score: Score) -> float:
        r"""Calculates weighted accuracy for an osu!mania score.

        :param score: The score to calculate accuracy for
        :type score: aiosu.models.score.Score
        :return: Weighted accuracy to be used in pp calculation
        :rtype: float
        """
        count_perfect = score.statistics.count_geki
        count_great = score.statistics.count_300
        count_good = score.statistics.count_katu
        count_ok = score.statistics.count_100
        count_meh = score.statistics.count_50
        count_miss = score.statistics.count_miss

        total_hits = (
            count_perfect + count_ok + count_great + count_good + count_meh + count_miss
        )

        if total_hits <= 0:
            return 0.0

        return max(
            (
                +(count_perfect * 320)
                + (count_great * 300)
                + (count_good * 200)
                + (count_ok * 100)
                + (count_meh * 50)
            )
            / (total_hits * 320),
            0.0,
        )


class CatchAccuracyCalculator(AbstractAccuracyCalculator):
    @staticmethod
    def calculate(score: Score) -> float:
        r"""Calculates accuracy for an osu!catch score.

        :param score: The score to calculate accuracy for
        :type score: aiosu.models.score.Score
        :return: Accuracy for the given score
        :rtype: float
        """
        fruits_hit = score.statistics.count_300
        ticks_hit = score.statistics.count_100
        tiny_ticks_hit = score.statistics.count_50
        tiny_ticks_missed = score.statistics.count_katu
        misses = score.statistics.count_miss

        total_hits = (
            tiny_ticks_hit + ticks_hit + fruits_hit + misses + tiny_ticks_missed
        )
        successful_hits = tiny_ticks_hit + ticks_hit + fruits_hit

        if total_hits <= 0:
            return 0.0

        return max(float(successful_hits) / total_hits, 0.0)

    @classmethod
    def calculate_weighted(cls, score: Score) -> float:
        r"""Calculates weighted accuracy for an osu!catch score.

        :param score: The score to calculate accuracy for
        :type score: aiosu.models.score.Score
        :return: Weighted accuracy to be used in pp calculation
        :rtype: float
        """
        return cls.calculate(score)



"""
This module contains models for User objects.
"""
from __future__ import annotations

from datetime import datetime
from enum import Enum
from enum import unique
from typing import Any
from typing import Literal
from typing import Optional
from typing import TYPE_CHECKING

from pydantic import computed_field
from pydantic import Field

from .base import BaseModel
from .common import Country
from .common import HTMLBody
from .common import TimestampedCount
from .gamemode import Gamemode

if TYPE_CHECKING:
    from typing import Callable

__all__ = (
    "User",
    "UserAccountHistory",
    "UserBadge",
    "UserGradeCounts",
    "UserGroup",
    "UserKudosu",
    "UserLevel",
    "UserProfileCover",
    "UserProfileTournamentBanner",
    "UserQueryType",
    "UserRankHistoryElement",
    "UserStatsVariant",
    "UserStats",
    "UserAccountHistoryType",
    "UserRankHighest",
    "ManiaStatsVariantsType",
)


cast_int: Callable[..., int] = lambda x: int(x or 0)
cast_float: Callable[..., float] = lambda x: float(x or 0)

UserAccountHistoryType = Literal[
    "note",
    "restriction",
    "silence",
    "tournament_ban",
]

ManiaStatsVariantsType = Literal[
    "4k",
    "7k",
]

OLD_QUERY_TYPES = {
    "ID": "id",
    "USERNAME": "string",
}


@unique
class UserQueryType(Enum):
    ID = "id"
    USERNAME = "username"

    @computed_field  # type: ignore
    @property
    def old_api_name(self) -> str:
        return OLD_QUERY_TYPES[self.name]

    @computed_field  # type: ignore
    @property
    def new_api_name(self) -> str:
        return self.value

    @classmethod
    def _missing_(cls, query: object) -> Any:
        if isinstance(query, str):
            query = query.lower()
        for q in list(UserQueryType):
            if query in (q.old_api_name, q.new_api_name):
                return q
        raise ValueError(f"UserQueryType {query} does not exist.")


class UserLevel(BaseModel):
    current: int
    progress: float

    @classmethod
    def _from_api_v1(cls, data: Any) -> UserLevel:
        level = cast_float(data["level"])
        current = int(level)
        progress = (level - current) * 100
        return cls.model_validate({"current": current, "progress": progress})


class UserKudosu(BaseModel):
    total: int
    available: int


class UserRankHistoryElement(BaseModel):
    mode: str
    data: list[int]

    @computed_field  # type: ignore
    @property
    def average_gain(self) -> float:
        r"""Average rank gain.

        :return: Average rank gain for a user
        :rtype: float
        """
        return (self.data[1] - self.data[-1]) / len(self.data)


class UserRankHighest(BaseModel):
    rank: int
    updated_at: datetime


class UserProfileCover(BaseModel):
    url: str
    custom_url: Optional[str] = None
    id: Optional[str] = None


class UserProfileTournamentBanner(BaseModel):
    tournament_id: int
    id: Optional[int] = None
    image: Optional[str] = None
    image_2_x: Optional[str] = Field(default=None, alias="image@2x")


class UserBadge(BaseModel):
    awarded_at: datetime
    description: str
    image_url: str
    url: str


class UserAccountHistory(BaseModel):
    id: int
    timestamp: datetime
    length: int
    permanent: bool
    type: UserAccountHistoryType
    description: Optional[str] = None


class UserGradeCounts(BaseModel):
    ssh: Optional[int] = None
    """Number of Silver SS ranks achieved."""
    ss: Optional[int] = None
    """Number of SS ranks achieved."""
    sh: Optional[int] = None
    """Number of Silver S ranks achieved."""
    s: Optional[int] = None
    """Number of S ranks achieved."""
    a: Optional[int] = None
    """Number of A ranks achieved."""

    @classmethod
    def _from_api_v1(cls, data: Any) -> UserGradeCounts:
        return cls.model_validate(
            {
                "ss": cast_int(data["count_rank_ss"]),
                "ssh": cast_int(data["count_rank_ssh"]),
                "s": cast_int(data["count_rank_s"]),
                "sh": cast_int(data["count_rank_sh"]),
                "a": cast_int(data["count_rank_a"]),
            },
        )


class UserGroup(BaseModel):
    id: int
    identifier: str
    name: str
    short_name: str
    has_listing: bool
    has_playmodes: bool
    is_probationary: bool
    colour: Optional[str] = None
    playmodes: Optional[list[Gamemode]] = None
    description: Optional[str] = None


class UserStatsVariant(BaseModel):
    mode: Gamemode
    variant: str
    pp: float
    country_rank: Optional[int] = None
    global_rank: Optional[int] = None


class UserStats(BaseModel):
    """Fields are marked as optional since they might be missing from rankings other than performance."""

    ranked_score: Optional[int] = None
    play_count: Optional[int] = None
    grade_counts: Optional[UserGradeCounts] = None
    total_hits: Optional[int] = None
    is_ranked: Optional[bool] = None
    total_score: Optional[int] = None
    level: Optional[UserLevel] = None
    hit_accuracy: Optional[float] = None
    play_time: Optional[int] = None
    pp: Optional[float] = None
    pp_exp: Optional[float] = None
    replays_watched_by_others: Optional[int] = None
    maximum_combo: Optional[int] = None
    global_rank: Optional[int] = None
    global_rank_exp: Optional[int] = None
    country_rank: Optional[int] = None
    user: Optional[User] = None
    count_300: Optional[int] = None
    count_100: Optional[int] = None
    count_50: Optional[int] = None
    count_miss: Optional[int] = None
    variants: Optional[list[UserStatsVariant]] = None

    @computed_field  # type: ignore
    @property
    def pp_per_playtime(self) -> float:
        r"""PP per playtime.

        :return: PP per playtime
        :rtype: float
        """
        if not self.play_time or not self.pp:
            return 0
        return self.pp / self.play_time * 3600

    @classmethod
    def _from_api_v1(cls, data: Any) -> UserStats:
        """Some fields can be None, we want to force them to cast to a value."""
        return cls.model_validate(
            {
                "level": UserLevel._from_api_v1(data),
                "pp": cast_float(data["pp_raw"]),
                "global_rank": cast_int(data["pp_rank"]),
                "country_rank": cast_int(data["pp_country_rank"]),
                "ranked_score": cast_int(data["ranked_score"]),
                "hit_accuracy": cast_float(data["accuracy"]),
                "play_count": cast_int(data["playcount"]),
                "play_time": cast_int(data["total_seconds_played"]),
                "total_score": cast_int(data["total_score"]),
                "total_hits": cast_int(data["count300"])
                + cast_int(data["count100"])
                + cast_int(data["count50"]),
                "is_ranked": cast_float(data["pp_raw"]) != 0,
                "grade_counts": UserGradeCounts._from_api_v1(data),
                "count_300": cast_int(data["count300"]),
                "count_100": cast_int(data["count100"]),
                "count_50": cast_int(data["count50"]),
            },
        )


class UserAchievmement(BaseModel):
    achieved_at: datetime
    achievement_id: int


class User(BaseModel):
    avatar_url: str
    country_code: str
    id: int
    username: str
    default_group: Optional[str] = None
    is_active: Optional[bool] = None
    is_bot: Optional[bool] = None
    is_online: Optional[bool] = None
    is_supporter: Optional[bool] = None
    pm_friends_only: Optional[bool] = None
    profile_colour: Optional[str] = None
    is_deleted: Optional[bool] = None
    last_visit: Optional[datetime] = None
    discord: Optional[str] = None
    has_supported: Optional[bool] = None
    interests: Optional[str] = None
    join_date: Optional[datetime] = None
    kudosu: Optional[UserKudosu] = None
    location: Optional[str] = None
    max_blocks: Optional[int] = None
    max_friends: Optional[int] = None
    occupation: Optional[str] = None
    playmode: Optional[Gamemode] = None
    playstyle: Optional[list[str]] = None
    post_count: Optional[int] = None
    profile_order: Optional[list[str]] = None
    title: Optional[str] = None
    twitter: Optional[str] = None
    website: Optional[str] = None
    country: Optional[Country] = None
    cover: Optional[UserProfileCover] = None
    is_restricted: Optional[bool] = None
    account_history: Optional[list[UserAccountHistory]] = None
    active_tournament_banner: Optional[UserProfileTournamentBanner] = None
    badges: Optional[list[UserBadge]] = None
    beatmap_playcounts_count: Optional[int] = None
    favourite_beatmapset_count: Optional[int] = None
    follower_count: Optional[int] = None
    graveyard_beatmapset_count: Optional[int] = None
    groups: Optional[list[UserGroup]] = None
    loved_beatmapset_count: Optional[int] = None
    monthly_playcounts: Optional[list[TimestampedCount]] = None
    page: Optional[HTMLBody] = None
    pending_beatmapset_count: Optional[int] = None
    previous_usernames: Optional[list[str]] = None
    ranked_beatmapset_count: Optional[int] = None
    replays_watched_counts: Optional[list[TimestampedCount]] = None
    scores_best_count: Optional[int] = None
    scores_first_count: Optional[int] = None
    scores_recent_count: Optional[int] = None
    statistics: Optional[UserStats] = None
    support_level: Optional[int] = None
    user_achievements: Optional[list[UserAchievmement]] = None
    rank_history: Optional[UserRankHistoryElement] = None
    rank_highest: Optional[UserRankHighest] = None

    @computed_field  # type: ignore
    @property
    def url(self) -> str:
        return f"https://osu.ppy.sh/users/{self.id}"

    @classmethod
    def _from_api_v1(cls, data: Any) -> User:
        return cls.model_validate(
            {
                "avatar_url": f"https://s.ppy.sh/a/{data['user_id']}",
                "country_code": data["country"],
                "id": data["user_id"],
                "username": data["username"],
                "join_date": data["join_date"],
                "statistics": UserStats._from_api_v1(data),
            },
        )


UserStats.model_rebuild()



"""
This module contains models for Score objects.
"""
from __future__ import annotations

from datetime import datetime
from typing import Optional
from typing import TYPE_CHECKING

from pydantic import computed_field
from pydantic import model_validator

from ..utils.accuracy import CatchAccuracyCalculator
from ..utils.accuracy import ManiaAccuracyCalculator
from ..utils.accuracy import OsuAccuracyCalculator
from ..utils.accuracy import TaikoAccuracyCalculator
from .base import BaseModel
from .beatmap import Beatmap
from .beatmap import Beatmapset
from .common import CurrentUserAttributes
from .gamemode import Gamemode
from .mods import Mods
from .user import User

if TYPE_CHECKING:
    from typing import Any
    from .. import v1

__all__ = (
    "Score",
    "ScoreStatistics",
    "ScoreWeight",
    "calculate_score_completion",
)

accuracy_calculators = {
    "osu": OsuAccuracyCalculator(),
    "mania": ManiaAccuracyCalculator(),
    "taiko": TaikoAccuracyCalculator(),
    "fruits": CatchAccuracyCalculator(),
}


def calculate_score_completion(
    mode: Gamemode,
    statistics: ScoreStatistics,
    beatmap: Beatmap,
) -> Optional[float]:
    """Calculates completion for a score.

    :param mode: The gamemode of the score
    :type mode: aiosu.models.gamemode.Gamemode
    :param statistics: The statistics of the score
    :type statistics: aiosu.models.score.ScoreStatistics
    :param beatmap: The beatmap of the score
    :type beatmap: aiosu.models.beatmap.Beatmap
    :raises ValueError: If the gamemode is unknown
    :return: Completion for the given score
    :rtype: Optional[float]
    """
    if not beatmap.count_objects:
        return None

    if mode == Gamemode.STANDARD:
        return (
            (
                statistics.count_300
                + statistics.count_100
                + statistics.count_50
                + statistics.count_miss
            )
            / beatmap.count_objects
        ) * 100
    elif mode == Gamemode.TAIKO:
        return (
            (statistics.count_300 + statistics.count_100 + statistics.count_miss)
            / beatmap.count_objects
        ) * 100
    elif mode == Gamemode.CTB:
        return (
            (statistics.count_300 + statistics.count_100 + +statistics.count_miss)
            / beatmap.count_objects
        ) * 100
    elif mode == Gamemode.MANIA:
        return (
            (
                statistics.count_300
                + statistics.count_100
                + statistics.count_50
                + statistics.count_miss
                + statistics.count_geki
                + statistics.count_katu
            )
            / beatmap.count_objects
        ) * 100

    raise ValueError("Unknown mode specified.")


class ScoreWeight(BaseModel):
    percentage: float
    pp: float


class ScoreStatistics(BaseModel):
    count_50: int
    count_100: int
    count_300: int
    count_miss: int
    count_geki: int
    count_katu: int

    @model_validator(mode="before")
    @classmethod
    def _convert_none_to_zero(cls, values: dict[str, Any]) -> dict[str, Any]:
        # Lazer API returns null for some statistics
        for key in values:
            if values[key] is None:
                values[key] = 0
        return values

    @classmethod
    def _from_api_v1(cls, data: Any) -> ScoreStatistics:
        return cls.model_validate(
            {
                "count_50": data["count50"],
                "count_100": data["count100"],
                "count_300": data["count300"],
                "count_geki": data["countgeki"],
                "count_katu": data["countkatu"],
                "count_miss": data["countmiss"],
            },
        )


class Score(BaseModel):
    user_id: int
    accuracy: float
    mods: Mods
    score: int
    max_combo: int
    passed: bool
    perfect: bool
    statistics: ScoreStatistics
    rank: str
    created_at: datetime
    mode: Gamemode
    replay: bool
    id: Optional[int] = None
    """Always present except for API v1 recent scores."""
    pp: Optional[float] = 0
    best_id: Optional[int] = None
    beatmap: Optional[Beatmap] = None
    beatmapset: Optional[Beatmapset] = None
    weight: Optional[ScoreWeight] = None
    user: Optional[User] = None
    rank_global: Optional[int] = None
    rank_country: Optional[int] = None
    type: Optional[str] = None
    current_user_attributes: Optional[CurrentUserAttributes] = None
    beatmap_id: Optional[int] = None
    """Only present on API v1"""

    @computed_field  # type: ignore
    @property
    def completion(self) -> Optional[float]:
        """Beatmap completion.

        :raises ValueError: If mode is unknown
        :return: Beatmap completion of a score (%). 100% for passes. None if no beatmap.
        :rtype: Optional[float]
        """
        if not self.beatmap:
            return None

        if self.passed:
            return 100.0

        return calculate_score_completion(self.mode, self.statistics, self.beatmap)

    @computed_field  # type: ignore
    @property
    def score_url(self) -> Optional[str]:
        r"""Link to the score.

        :return: Link to the score on the osu! website
        :rtype: Optional[str]
        """
        if (not self.id and not self.best_id) or not self.passed:
            return None
        return (
            f"https://osu.ppy.sh/scores/{self.mode.name_api}/{self.best_id}"
            if self.best_id
            else f"https://osu.ppy.sh/scores/{self.id}"
        )

    @computed_field  # type: ignore
    @property
    def replay_url(self) -> Optional[str]:
        r"""Link to the replay.

        :return: Link to download the replay on the osu! website
        :rtype: Optional[str]
        """
        if not self.replay:
            return None
        return (
            f"https://osu.ppy.sh/scores/{self.mode.name_api}/{self.best_id}/download"
            if self.best_id
            else f"https://osu.ppy.sh/scores/{self.id}/download"
        )

    @model_validator(mode="before")
    @classmethod
    def _fail_rank(cls, values: dict[str, Any]) -> dict[str, Any]:
        if not values["passed"]:
            values["rank"] = "F"
        return values

    async def request_beatmap(self, client: v1.Client) -> None:
        r"""For v1 Scores: requests the beatmap from the API and sets it.

        :param client: An API v1 Client
        :type client: aiosu.v1.client.Client
        """
        if self.beatmap_id is None:
            raise ValueError("Score has unknown beatmap ID")
        if self.beatmap is None and self.beatmapset is None:
            sets = await client.get_beatmap(
                mode=self.mode,
                beatmap_id=self.beatmap_id,
            )
            self.beatmapset = sets[0]
            self.beatmap = sets[0].beatmaps[0]  # type: ignore

    @classmethod
    def _from_api_v1(
        cls,
        data: Any,
        mode: Gamemode,
    ) -> Score:
        statistics = ScoreStatistics._from_api_v1(data)
        score = cls.model_validate(
            {
                "id": data["score_id"],
                "user_id": data["user_id"],
                "accuracy": 0.0,
                "mods": int(data["enabled_mods"]),
                "score": data["score"],
                "pp": data.get("pp", 0.0),
                "max_combo": data["maxcombo"],
                "passed": data["rank"] != "F",
                "perfect": data["perfect"],
                "statistics": statistics,
                "rank": data["rank"],
                "created_at": data["date"],
                "mode": mode,
                "beatmap_id": data.get("beatmap_id"),
                "replay": data.get("replay_available", False),
            },
        )
        score.accuracy = accuracy_calculators[str(mode)].calculate(score)
        return score



"""
This module contains base models for objects.
"""
from __future__ import annotations

import pydantic
from pydantic import ConfigDict

__all__ = (
    "BaseModel",
    "FrozenModel",
)


class BaseModel(pydantic.BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    @classmethod
    def model_validate_file(cls, path: str) -> BaseModel:
        """Validates a model from a file.

        :param path: The path to the file
        :type path: str
        :raises TypeError: If the file is not a JSON file
        :return: The validated model
        :rtype: aiosu.models.base.BaseModel
        """
        with open(path) as f:
            return cls.model_validate_json(f.read())


class FrozenModel(BaseModel):
    model_config = ConfigDict(
        arbitrary_types_allowed=True,
        populate_by_name=True,
        frozen=True,
    )



"""
This module contains models for API events.
"""
from __future__ import annotations

from datetime import datetime
from typing import Literal
from typing import Optional
from typing import TYPE_CHECKING

from pydantic import Field

from .base import BaseModel
from .beatmap import BeatmapRankStatus
from .common import Achievement
from .gamemode import Gamemode

if TYPE_CHECKING:
    pass

__all__ = (
    "Event",
    "EventBeatmap",
    "EventBeatmapset",
    "EventUser",
    "EventType",
)

EventType = Literal[
    "achievement",
    "beatmapPlaycount",
    "beatmapsetApprove",
    "beatmapsetDelete",
    "beatmapsetRevive",
    "beatmapsetUpdate",
    "beatmapsetUpload",
    "rank",
    "rankLost",
    "usernameChange",
    "userSupportAgain",
    "userSupportFirst",
    "userSupportGift",
]


class EventBeatmap(BaseModel):
    title: str
    url: str


class EventBeatmapset(BaseModel):
    title: str
    url: str


class EventUser(BaseModel):
    username: str
    url: str
    previous_username: Optional[str] = Field(default=None, alias="previousUsername")


class Event(BaseModel):
    created_at: datetime
    id: int
    type: EventType
    r"""Information on types: https://github.com/ppy/osu-web/blob/master/resources/assets/lib/interfaces/event-json.ts"""
    parse_error: Optional[bool] = None
    achievment: Optional[Achievement] = None
    user: Optional[EventUser] = None
    beatmap: Optional[EventBeatmap] = None
    beatmapset: Optional[EventBeatmapset] = None
    approval: Optional[BeatmapRankStatus] = None
    count: Optional[int] = None
    rank: Optional[int] = None
    mode: Optional[Gamemode] = None
    score_rank: Optional[str] = Field(default=None, alias="scoreRank")



"""
This module contains models for rankings.
"""
from __future__ import annotations

from typing import Literal
from typing import Optional

from .beatmap import Beatmapset
from .common import CursorModel
from .spotlight import Spotlight
from .user import UserStats


__all__ = (
    "RankingFilter",
    "RankingVariant",
    "RankingType",
    "Rankings",
)

RankingFilter = Literal["all", "friends"]
RankingType = Literal["performance", "score", "country", "charts"]
RankingVariant = Literal["4k", "7k"]


class Rankings(CursorModel):
    ranking: list[UserStats]
    total: Optional[int] = None
    spotlight: Optional[Spotlight] = None
    beatmapsets: Optional[list[Beatmapset]] = None



"""
This module contains models for featured artists.
"""
from __future__ import annotations

from datetime import datetime
from typing import Optional

from pydantic import Field

from .base import BaseModel
from .common import CursorModel

__all__ = (
    "ArtistTrack",
    "ArtistResponse",
)


class ArtistSearch(BaseModel):
    is_default_sort: bool
    sort: str


class Artist(BaseModel):
    id: int
    name: str


class Album(BaseModel):
    id: int
    artist_id: int
    title: str
    title_romanized: str
    genre: str
    is_new: bool
    cover_url: str


class ArtistTrack(BaseModel):
    id: int
    artist_id: int
    bpm: float
    cover_url: str
    exclusive: bool
    genre: str
    is_new: bool
    length: str
    title: str
    artist: Artist
    osz_url: str = Field(alias="osz")
    preview_url: str = Field(alias="preview")
    album: Optional[Album] = None
    album_id: Optional[int] = None
    updated_at: Optional[datetime] = None
    version: Optional[str] = None


class ArtistResponse(CursorModel):
    """Artist response model."""

    artist_tracks: list[ArtistTrack]
    search: Optional[ArtistSearch] = None



"""
This module contains a collection of utility functions.
"""
from __future__ import annotations

...  # disable import sorting

from . import auth
from . import accuracy
from . import binary
from . import performance
from . import replay



"""
This module contains models for miscellaneous objects.
"""
from __future__ import annotations

from datetime import datetime
from functools import partial
from typing import Any
from typing import Coroutine
from typing import Literal
from typing import Optional

from emojiflags.lookup import lookup as flag_lookup  # type: ignore
from pydantic import computed_field
from pydantic import Field
from pydantic import field_validator

from .base import BaseModel
from .gamemode import Gamemode

__all__ = (
    "Achievement",
    "Country",
    "CurrentUserAttributes",
    "TimestampedCount",
    "CursorModel",
    "SortTypes",
    "HTMLBody",
)


SortTypes = Literal["id_asc", "id_desc"]


class TimestampedCount(BaseModel):
    start_date: datetime
    count: int

    @field_validator("start_date", mode="before")
    @classmethod
    def _date_validate(cls, v: str) -> datetime:
        return datetime.strptime(v, "%Y-%m-%d")


class Achievement(BaseModel):
    id: int
    name: str
    slug: str
    desciption: str
    grouping: str
    icon_url: str
    mode: Gamemode
    ordering: int
    instructions: Optional[str] = None


class Country(BaseModel):
    code: str
    name: str

    @computed_field  # type: ignore
    @property
    def flag_emoji(self) -> str:
        r"""Emoji for the flag.

        :return: Unicode emoji representation of the country's flag
        :rtype: str
        """
        return flag_lookup(self.code)


class HTMLBody(BaseModel):
    html: str
    raw: Optional[str] = None
    bbcode: Optional[str] = None


class PinAttributes(BaseModel):
    is_pinned: bool
    score_id: int
    score_type: str


class CurrentUserAttributes(BaseModel):
    can_beatmap_update_owner: Optional[bool] = None
    can_delete: Optional[bool] = None
    can_edit_metadata: Optional[bool] = None
    can_edit_tags: Optional[bool] = None
    can_hype: Optional[bool] = None
    can_hype_reason: Optional[str] = None
    can_love: Optional[bool] = None
    can_remove_from_loved: Optional[bool] = None
    is_watching: Optional[bool] = None
    new_hype_time: Optional[datetime] = None
    nomination_modes: Optional[list[Gamemode]] = None
    remaining_hype: Optional[int] = None
    can_destroy: Optional[bool] = None
    can_reopen: Optional[bool] = None
    can_moderate_kudosu: Optional[bool] = None
    can_resolve: Optional[bool] = None
    vote_score: Optional[int] = None
    can_message: Optional[bool] = None
    can_message_error: Optional[str] = None
    last_read_id: Optional[int] = None
    can_new_comment: Optional[bool] = None
    can_new_comment_reason: Optional[str] = None
    pin: Optional[PinAttributes] = None


class CursorModel(BaseModel):
    r"""NOTE: This model is not serializable by orjson directly.

    Use the provided .model_dump_json() or .model_dump() methods instead.
    """

    cursor_string: Optional[str] = None
    next: Optional[partial[Coroutine[Any, Any, CursorModel]]] = Field(
        default=None,
        exclude=True,
    )
    """Partial function to get the next page of results."""



"""
This module contains the Forum model.
"""
from __future__ import annotations

from datetime import datetime
from typing import Literal
from typing import Optional

from .base import BaseModel
from .common import CursorModel
from .common import HTMLBody

__all__ = [
    "ForumPost",
    "ForumTopic",
    "ForumTopicResponse",
    "ForumTopicType",
    "ForumPoll",
    "ForumPollOption",
    "ForumCreateTopicResponse",
]

ForumTopicType = Literal[
    "normal",
    "sticky",
    "announcement",
]


class ForumPollOption(BaseModel):
    id: int
    text: HTMLBody
    vote_count: Optional[int] = None


class ForumPoll(BaseModel):
    allow_vote_change: bool
    hide_incomplete_results: bool
    max_votes: int
    total_vote_count: int
    options: list[ForumPollOption]
    started_at: datetime
    title: HTMLBody
    ended_at: Optional[datetime] = None
    last_vote_at: Optional[datetime] = None


class ForumTopic(BaseModel):
    id: int
    title: str
    created_at: datetime
    first_post_id: int
    last_post_id: int
    forum_id: int
    is_locked: bool
    post_count: int
    type: ForumTopicType
    updated_at: datetime
    user_id: int
    deleted_at: Optional[datetime] = None
    poll: Optional[ForumPoll] = None


class ForumPost(BaseModel):
    id: int
    created_at: datetime
    forum_id: int
    topic_id: int
    user_id: int
    edited_by_id: Optional[int] = None
    edited_at: Optional[datetime] = None
    deleted_at: Optional[datetime] = None
    body: Optional[HTMLBody] = None


class ForumTopicResponse(CursorModel):
    topic: ForumTopic
    posts: list[ForumPost]


class ForumCreateTopicResponse(BaseModel):
    topic: ForumTopic
    post: ForumPost



"""
This module contains models for lazer specific data.
"""
from __future__ import annotations

from datetime import datetime
from typing import Any
from typing import Optional

from pydantic import computed_field
from pydantic import Field
from pydantic import model_validator

from .base import BaseModel
from .beatmap import Beatmap
from .beatmap import Beatmapset
from .common import CurrentUserAttributes
from .gamemode import Gamemode
from .score import ScoreWeight
from .user import User

__all__ = (
    "LazerMod",
    "LazerScoreStatistics",
    "LazerReplayData",
    "LazerScore",
)


def calculate_score_completion(
    statistics: LazerScoreStatistics,
    beatmap: Beatmap,
) -> Optional[float]:
    """Calculates completion for a score.

    :param statistics: The statistics of the score
    :type statistics: aiosu.models.lazer.LazerScoreStatistics
    :param beatmap: The beatmap of the score
    :type beatmap: aiosu.models.beatmap.Beatmap
    :raises ValueError: If the gamemode is unknown
    :return: Completion for the given score
    :rtype: Optional[float]
    """
    if not beatmap.count_objects:
        return None

    return (
        (
            statistics.perfect
            + statistics.good
            + statistics.great
            + statistics.ok
            + statistics.meh
            + statistics.miss
        )
        / beatmap.count_objects
    ) * 100


class LazerMod(BaseModel):
    """Temporary model for lazer mods."""

    acronym: str
    settings: dict[str, Any] = Field(default_factory=dict)

    def __str__(self) -> str:
        return self.acronym


class LazerScoreStatistics(BaseModel):
    ok: int = 0
    meh: int = 0
    miss: int = 0
    great: int = 0
    ignore_hit: int = 0
    ignore_miss: int = 0
    large_bonus: int = 0
    large_tick_hit: int = 0
    large_tick_miss: int = 0
    small_bonus: int = 0
    small_tick_hit: int = 0
    small_tick_miss: int = 0
    good: int = 0
    perfect: int = 0
    legacy_combo_increase: int = 0

    @computed_field  # type: ignore
    @property
    def count_300(self) -> int:
        return self.great

    @computed_field  # type: ignore
    @property
    def count_100(self) -> int:
        return self.ok

    @computed_field  # type: ignore
    @property
    def count_50(self) -> int:
        return self.meh

    @computed_field  # type: ignore
    @property
    def count_miss(self) -> int:
        return self.miss

    @computed_field  # type: ignore
    @property
    def count_geki(self) -> int:
        return self.perfect

    @computed_field  # type: ignore
    @property
    def count_katu(self) -> int:
        return self.good


class LazerReplayData(BaseModel):
    mods: list[LazerMod]
    statistics: LazerScoreStatistics
    maximum_statistics: LazerScoreStatistics


class LazerScore(BaseModel):
    id: int
    accuracy: float
    beatmap_id: int
    max_combo: int
    maximum_statistics: LazerScoreStatistics
    mods: list[LazerMod]
    passed: bool
    rank: str
    ruleset_id: int
    ended_at: datetime
    statistics: LazerScoreStatistics
    total_score: int
    user_id: int
    replay: bool
    type: str
    current_user_attributes: CurrentUserAttributes
    beatmap: Beatmap
    beatmapset: Beatmapset
    user: User
    build_id: Optional[int] = None
    started_at: Optional[datetime] = None
    best_id: Optional[int] = None
    legacy_perfect: Optional[bool] = None
    pp: Optional[float] = None
    weight: Optional[ScoreWeight] = None

    @computed_field  # type: ignore
    @property
    def mods_str(self) -> str:
        return "".join(str(mod) for mod in self.mods)

    @computed_field  # type: ignore
    @property
    def created_at(self) -> datetime:
        return self.ended_at

    @computed_field  # type: ignore
    @property
    def completion(self) -> Optional[float]:
        """Beatmap completion.

        :return: Beatmap completion of a score (%). 100% for passes. None if no beatmap.
        :rtype: Optional[float]
        """
        if not self.beatmap:
            return None

        if self.passed:
            return 100.0

        return calculate_score_completion(self.statistics, self.beatmap)

    @computed_field  # type: ignore
    @property
    def mode(self) -> Gamemode:
        return Gamemode(self.ruleset_id)

    @computed_field  # type: ignore
    @property
    def score(self) -> int:
        return self.total_score

    @computed_field  # type: ignore
    @property
    def score_url(self) -> Optional[str]:
        r"""Link to the score.

        :return: Link to the score on the osu! website
        :rtype: Optional[str]
        """
        if (not self.id and not self.best_id) or not self.passed:
            return None
        return (
            f"https://osu.ppy.sh/scores/{self.mode.name_api}/{self.best_id}"
            if self.best_id
            else f"https://osu.ppy.sh/scores/{self.id}"
        )

    @computed_field  # type: ignore
    @property
    def replay_url(self) -> Optional[str]:
        r"""Link to the replay.

        :return: Link to download the replay on the osu! website
        :rtype: Optional[str]
        """
        if not self.replay:
            return None
        return (
            f"https://osu.ppy.sh/scores/{self.mode.name_api}/{self.best_id}/download"
            if self.best_id
            else f"https://osu.ppy.sh/scores/{self.id}/download"
        )

    @model_validator(mode="before")
    @classmethod
    def _fail_rank(cls, values: dict[str, Any]) -> dict[str, Any]:
        if not values["passed"]:
            values["rank"] = "F"
        return values
